# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref test for views."""

import json
import os
from datetime import date
from tempfile import NamedTemporaryFile

from yams.schema import role_name
from logilab.common.testlib import TestCase, unittest_main

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.rdfio import default_graph

from cubes.saem_ref.views import seda
from cubes.saem_ref.views.widgets import process_incomplete_date

import testutils


class SetDateTC(TestCase):

    def test_set_date_month_beginning(self):
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("5/6/1997"))
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("5/6/97"))
        self.assertEqual(date(2012, 6, 5), process_incomplete_date("5/6/12"))
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("1997/6/5"))
        self.assertEqual(date(1997, 6, 1), process_incomplete_date("6/1997"))
        self.assertEqual(date(1997, 1, 1), process_incomplete_date("1997"))

    def test_set_date_month_end(self):
        self.assertEqual(date(1994, 8, 28), process_incomplete_date("28/08/1994", True))
        self.assertEqual(date(1994, 8, 31), process_incomplete_date("08/1994", True))
        self.assertEqual(date(1994, 2, 28), process_incomplete_date("1994/02", True))
        self.assertEqual(date(1994, 12, 31), process_incomplete_date("1994", True))

    def test_set_date_failure(self):
        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("31/02/2012")
        self.assertIn("day is out of range for month", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("20/14/2015")
        self.assertIn("month must be in 1..12", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("20")


class FuncViewsTC(CubicWebTC):

    def test_eac_import_ok(self):
        regid = 'saem_ref.eac-import'
        fname = 'FRAD033_EAC_00001_simplified.xml'
        with self.admin_access.web_request() as req:
            anything = req.execute('Any X LIMIT 1')
            # simply test the form properly render and is well formed
            self.view(regid, rset=anything, req=req, template=None)
            fields = {'file': (fname, open(self.datapath('EAC/' + fname)))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            req.view(regid)
            rset = req.find('Agent', name=u'Gironde. Conseil général')
            self.assertTrue(rset)

    def test_eac_non_unique_isni(self):
        regid = 'saem_ref.eac-import'
        fname = 'FRAD033_EAC_00001_simplified.xml'
        # Query for agents not related to a CWUser
        agent_rql = 'Any X WHERE X is Agent, NOT EXISTS(X agent_user U)'
        with self.admin_access.client_cnx() as cnx:
            akind = cnx.find('AgentKind', name=u'person').one()
            # ISNI is the same as the agent in EAC file.
            cnx.create_entity('Agent', name=u'bob', isni=u'22330001300016',
                              agent_kind=akind)
            cnx.commit()
            rset = cnx.execute(agent_rql)
            self.assertEqual(len(rset), 1)
        with self.admin_access.web_request() as req:
            anything = req.execute('Any X LIMIT 1')
            # simply test the form properly render and is well formed
            self.view(regid, rset=anything, req=req, template=None)
            fields = {'file': (fname, open(self.datapath('EAC/' + fname)))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn('EAC import failed', html)
            # Still only one Agent.
            rset = req.execute(agent_rql)
            self.assertEqual(len(rset), 1)

    def test_eac_invalid_xml(self):
        regid = 'saem_ref.eac-import'
        fname = 'invalid_xml.xml'
        with self.admin_access.web_request() as req:
            fields = {'file': (fname, open(self.datapath('EAC/' + fname)))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn('Invalid XML file', html)

    def test_eac_missing_tag(self):
        regid = 'saem_ref.eac-import'
        fname = 'missing_tag.xml'
        with self.admin_access.web_request() as req:
            fields = {'file': (fname, open(self.datapath('EAC/' + fname)))}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn('Missing tag cpfDescription in XML file', html)

    def test_highlight_script_execution(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus')
            cnx.commit()
            url = scheme.absolute_url(highlight='toto')
        self.assertIn(
            '$(document).ready(function(){$("h1, h2, h3, h4, h5, table tbody td")'
            '.highlight("toto");});}',
            self.http_publish(url)[0])

    def test_highlight_on_rql_plain_text_search_same_etype(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus')
            concept1 = cnx.create_entity('Concept', in_scheme=scheme, definition=u'toto tata')
            concept2 = cnx.create_entity('Concept', in_scheme=scheme, definition=u'titi toto')
            cnx.create_entity('Label', label=u'toto', language_code=u'fr', label_of=concept1)
            cnx.create_entity('Label', label=u'tutu', language_code=u'fr', label_of=concept2)
            cnx.commit()
        url = 'http://testing.fr/cubicweb/view?rql=toto&__fromsearchbox=1&subvid=tsearch'
        html = self.http_publish(url)[0]
        self.assertIn(
            '<a href="http://testing.fr/cubicweb/%s?highlight=toto" title="">' % concept1.eid, html)
        self.assertIn(
            '<a href="http://testing.fr/cubicweb/%s?highlight=toto" title="">' % concept2.eid, html)

    def test_highlight_on_rql_plain_text_search(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus')
            concept = cnx.create_entity('Concept', in_scheme=scheme, definition=u'toto tata')
            akind = cnx.find('AgentKind', name=u'person').one()
            agent = cnx.create_entity('Agent', name=u'toto', isni=u'1234', agent_kind=akind)
            cnx.create_entity('Label', label=u'tutu', language_code=u'fr', label_of=concept)
            cnx.create_entity('Label', label=u'toti', language_code=u'fr', label_of=concept,
                              kind=u'alternative')
            cnx.commit()
        url = 'http://testing.fr/cubicweb/view?rql=toto&__fromsearchbox=1&subvid=tsearch'
        html = self.http_publish(url)[0]
        self.assertIn(
            '<a href="http://testing.fr/cubicweb/%s?highlight=toto" title="">' % concept.eid, html)
        self.assertIn(
            '<a href="http://testing.fr/cubicweb/%s?highlight=toto" title="">' % agent.eid, html)

    def test_skos_negociation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'musique')
            scheme.add_concept(u'pop')
            cnx.commit()
        with self.admin_access.web_request(headers={'Accept': 'application/rdf+xml'}) as req:
            result = self.app_handle_request(req, 'conceptscheme')
            with NamedTemporaryFile(delete=False) as fobj:
                try:
                    fobj.write(result)
                    fobj.close()
                    graph = default_graph()
                    graph.load('file://' + fobj.name, rdf_format='xml')
                finally:
                    os.unlink(fobj.name)

    def test_seda_document_type_code_vocabulary(self):
        """Check that we get correct values in the combo box for document type code."""
        with self.admin_access.web_request() as req:
            doc = self.vreg['etypes'].etype_class('SEDADocumentTypeCode')(req)
            form = self.vreg['forms'].select('edition', req, entity=doc)
            field = form.field_by_name('seda_document_type_code_value', 'subject')
            labels = [label for label, _ in field.vocabulary(form)]
            self.assertEqual(labels, [u'CDO', u'PDI', u'PDICTX', u'PDIFIX', u'PDIPRO', u'PDIREF',
                                      u'RI', u'RISEM', u'RISTR'])

    def test_seda_file_type_code_vocabulary(self):
        """Check that we get correct values in the combo box for file type code."""
        with self.admin_access.web_request() as req:
            doc = self.vreg['etypes'].etype_class('SEDAFileTypeCode')(req)
            form = self.vreg['forms'].select('edition', req, entity=doc)
            field = form.field_by_name('seda_file_type_code_value', 'subject')
            values = list(field.vocabulary(form))
            labels = [label for label, _ in values]
            self.assertEqual(len(labels), 367)
            self.assertIn(u'3GPP Audio/Video File 3GPP Audio/Video File', labels)
            self.assertIn(u'ESRI World File Format ESRI World File Format', labels)
            self.assertIn(u'pulse EKKO header file pulse EKKO header file', labels)

    def test_seda_character_set_code_vocabulary(self):
        """Check that we get correct values in the combo box for character set code."""
        with self.admin_access.web_request() as req:
            doc = self.vreg['etypes'].etype_class('SEDACharacterSetCode')(req)
            form = self.vreg['forms'].select('edition', req, entity=doc)
            field = form.field_by_name('seda_character_set_code_value', 'subject')
            values = list(field.vocabulary(form))
            labels = [label for label, _ in values]
            self.assertEqual(len(labels), 256)
            self.assertIn(u'ANSI_X3.110-1983 [RFC1345,KXS2]', labels)
            self.assertIn(u'windows-874', labels)
            self.assertIn(u'UTF-8 [RFC3629]', labels)

    def test_seda_get_related_version(self):
        """Check that we get correct results when asking for `draft`, `published`, `replaced`
        version of a profile."""
        with self.admin_access.web_request() as req:
            profile1 = testutils.publishable_profile(req, title=u'Profile 1')
            req.cnx.commit()
            profile1.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile1.cw_clear_all_caches()
            profile2 = testutils.publishable_profile(req, title=u'Profile 2', seda_replace=profile1)
            req.cnx.commit()
            profile2.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile2.cw_clear_all_caches()
            profile3 = testutils.publishable_profile(req, title=u'Profile 3', seda_replace=profile2)
            req.cnx.commit()
            profile3.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile3.cw_clear_all_caches()
            profile4 = testutils.publishable_profile(req, title=u'Profile 4', seda_replace=profile3)
            req.cnx.commit()

            def unwrap_generator(gen):
                try:
                    return next(iter(gen))
                except StopIteration:
                    return None

            # Draft profile
            box4 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile4)
            self.assertEqual(unwrap_generator(box4.predecessor()).eid, profile3.eid)
            self.assertIsNone(unwrap_generator(box4.current_version(state=u'published')))
            self.assertIsNone(unwrap_generator(box4.current_version(state=u'draft')))
            # Published profile
            box3 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile3)
            self.assertEqual(unwrap_generator(box3.predecessor()).eid, profile2.eid)
            self.assertIsNone(unwrap_generator(box3.current_version(state=u'published')))
            self.assertEqual(unwrap_generator(box3.current_version(state=u'draft')).eid,
                             profile4.eid)
            # Deprecated profile
            box2 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile2)
            self.assertIsNone(unwrap_generator(box2.predecessor()))
            self.assertEqual(unwrap_generator(box2.current_version(state=u'published')).eid,
                             profile3.eid)
            self.assertEqual(unwrap_generator(box2.current_version(state=u'draft')).eid,
                             profile4.eid)
            # Older deprecated profile
            box1 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile1)
            self.assertIsNone(unwrap_generator(box1.predecessor()))
            self.assertEqual(unwrap_generator(box1.current_version(state=u'published')).eid,
                             profile3.eid)
            self.assertEqual(unwrap_generator(box1.current_version(state=u'draft')).eid,
                             profile4.eid)

    def test_seda_actions_notype_to_import(self):
        with self.admin_access.web_request() as req:
            profile = testutils.setup_seda_profile(req)
            req.cnx.commit()
            profile._cw = req  # XXX
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertIn(seda.SEDAImportProfileArchiveObjectAction, actions['moreactions'])
            self.assertNotIn(seda.SEDAImportProfileDocumentAction, actions['moreactions'])
            actions = self.pactionsdict(req, profile.archives[0].as_rset())
            self.assertIn(seda.SEDAImportProfileArchiveObjectAction, actions['moreactions'])
            self.assertIn(seda.SEDAImportProfileDocumentAction, actions['moreactions'])
        with self.new_access('anon').web_request() as req:
            profile = req.entity_from_eid(profile.eid)
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertNotIn(seda.SEDAImportProfileArchiveObjectAction, actions['moreactions'])
            self.assertNotIn(seda.SEDAImportProfileDocumentAction, actions['moreactions'])

    def test_seda_actions_not_exportable(self):
        with self.admin_access.web_request() as req:
            profile = req.create_entity('SEDAProfile')
            req.cnx.commit()
            profile._cw = req  # XXX
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertNotIn(seda.SEDA1DownloadAction, actions['moreactions'])
            self.assertNotIn(seda.SEDA02DownloadAction, actions['moreactions'])
            req.create_entity('ProfileArchiveObject',
                              seda_name=req.create_entity('SEDAName'),
                              seda_parent=profile)
            req.cnx.commit()
            profile.cw_clear_all_caches()
        with self.admin_access.web_request() as req:
            profile = req.entity_from_eid(profile.eid)
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertIn(seda.SEDA1DownloadAction, actions['moreactions'])
            self.assertIn(seda.SEDA02DownloadAction, actions['moreactions'])

    def test_ark_agent_creation(self):
        with self.admin_access.web_request() as req:
            akind = req.cnx.find('AgentKind', name=u'person').one()
            agent = self.vreg['etypes'].etype_class('Agent')(req)
            agent.eid = 'A'
            fields = {role_name('name', 'subject'): u'007',
                      role_name('ark', 'subject'): u'',
                      role_name('agent_kind', 'subject'): str(akind.eid)}
            req.form = self.fake_form('edition', entity_field_dicts=[(agent, fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            agent = req.cnx.entity_from_eid(eid)
            self.assertEqual(agent.ark, u'saemref-test/%09d' % eid)

    def test_ark_scheme_creation(self):
        with self.admin_access.web_request() as req:
            scheme = self.vreg['etypes'].etype_class('ConceptScheme')(req)
            scheme.eid = 'A'
            fields = {role_name('ark', 'subject'): u''}
            req.form = self.fake_form('edition', entity_field_dicts=[(scheme, fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            scheme = req.cnx.entity_from_eid(eid)
            self.assertEqual(scheme.ark, u'saemref-test/%09d' % eid)
            concept = self.vreg['etypes'].etype_class('Concept')(req)
            concept.eid = 'A'
            concept_fields = {role_name('in_scheme', 'subject'): str(scheme.eid),
                              role_name('ark', 'subject'): u''}
            label = self.vreg['etypes'].etype_class('Label')(req)
            label.eid = 'B'
            label_fields = {role_name('label', 'subject'): u'Hello',
                            role_name('label_of', 'subject'): 'A'}
            req.form = self.fake_form('edition', entity_field_dicts=[(concept, concept_fields),
                                                                     (label, label_fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            concept = req.cnx.entity_from_eid(eid)
            self.assertEqual(concept.ark, u'saemref-test/%09d' % eid)

    def test_ark_url_rewrite(self):
        with self.admin_access.web_request() as req:
            rewriter = self.vreg['urlrewriting'].select('schemabased', req)
            _pmid, rset = rewriter.rewrite(req, u'/ark:/JoE/Dalton')
            self.assertEqual(rset.printable_rql(), 'Any X WHERE X ark "JoE/Dalton"')

    def test_ark_ws(self):
        with self.admin_access.web_request(headers={'Accept': 'application/json'},
                                           method='POST') as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'ark': 'saemref-test/ext-000000001'}])
        with self.admin_access.web_request(headers={'Accept': 'application/json'}) as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'This service is only accessible using POST.'}])
        with self.new_access('anon').web_request(headers={'Accept': 'application/json'},
                                                 method='POST') as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'This service requires authentication.'}])

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



from cubes.saem_ref import workflows


for etype in ('Agent', 'ConceptScheme', 'SEDAProfile'):
    make_workflowable(etype)
    workflows.define_publication_workflow(add_workflow, etype)
    rql('SET X in_state ST WHERE X is {etype}, ST name "draft", ST state_of W, '
        'W workflow_of ET, ET name "{etype}"'.format(etype=etype))
    commit()
    sync_schema_props_perms(etype)

sync_schema_props_perms('phone_number', syncperms=False)
sync_schema_props_perms('postal_address', syncperms=False)

sync_schema_props_perms('seda_character_set_code', syncperms=False)
sync_schema_props_perms('seda_document_type_code', syncperms=False)
sync_schema_props_perms('seda_file_type_code', syncperms=False)


for agent in rql('Any X,A WHERE X is Agent, X ark A, NOT X cwuri ~= "%ark:%"').entities():
    agent.cw_set(cwuri=agent.cw_adapt_to('IARKGenerator').build_url(agent.ark))
commit()


add_relation_type('archival_agent')
add_relation_type('related_concept_scheme')

rql('SET X archival_agent Y WHERE P seda_transferring_agent X, P seda_archival_agent Y')
drop_relation_type('seda_archival_agent')

for agent in rql('Any X WHERE X is Agent, X archival_role AR, AR name "deposit", NOT X archival_agent Y').entities():
    msg = 'You should add an archival agent to agent %s (%s)' % (agent.name, agent.absolute_url())
    print msg.encode('utf-8')



# add a new column to appears table
if repo.system_source.dbdriver == 'postgres':
    sql('ALTER TABLE appears ADD pwords text ARRAY;')
    sql('ALTER TABLE words ADD count INT DEFAULT 1;')
    # remove the old index
    sql('DROP INDEX words_unique_idx;')
    # add the unique index
    sql('CREATE UNIQUE INDEX words_unique_idx ON words (etype, word);')



# SEDA profiles now have an ark identifier in the schema (replacing identifier)
add_attribute('SEDAProfile', 'ark')
drop_attribute('SEDAProfile', 'identifier')

# Add an ark identifier to existing SEDA profiles
for profile in rql('SEDAProfile P').entities():
    ark = profile.cw_adapt_to('IARKGenerator').generate_ark()
    profile.cw_set(ark=ark, cwuri=u'ark:/' + ark)
commit()

sync_schema_props_perms('AgentFunction', syncperms=False)

sync_schema_props_perms('seda_transferring_agent', syncperms=False)
sync_schema_props_perms('seda_parent', syncperms=False)


sql("SET TIME ZONE 'Europe/Paris'")

for entity in schema.entities():
    if entity.final:
        continue
    change_attribute_type(entity.type, 'creation_date', 'TZDatetime', ask_confirm=False)
    change_attribute_type(entity.type, 'modification_date', 'TZDatetime', ask_confirm=False)

create_entity('ArchivalRole', name=u'seda-actor', commit=True)

add_attribute('SEDAProfile', 'support_seda_exports')
rql('SET X support_seda_exports "SEDA-0.2.xsd, SEDA-1.0.xsd"')
commit()

profile_wf = get_workflow_for('SEDAProfile')
publish = profile_wf.transition_by_name('publish')
publish.set_permissions(conditions=('U in_group G, G name IN ("users", "managers"),'
                                    'X support_seda_exports ~= "%SEDA-0.2%"',),
                        reset=True)
commit()

# 0.8 migration missed some permission changes
sync_schema_props_perms(syncprops=False)

# New relation replace between profiles
add_relation_type('seda_replace')

for stmt in repo.system_source.dbhelper.sql_create_sequence('ext_ark_count').split(';'):
    if stmt:
        sql(stmt)




if versions_map['saem_ref'][0] < (0, 7, 99):
    print "Migration has to be interrupted but is not finished."
    print "You should restart the command to fulfill it."
    import sys
    sys.exit(0)

repo.system_source.do_fti = False

for rtype in ('file_type_code', 'character_set_code', 'document_type_code'):
    cs = rql('Any CS WHERE CS scheme_relation RT, RT name %(rt)s', {'rt': rtype})[0][0]
    with cnx.deny_all_hooks_but('syncschema', 'metadata'):
        rename_relation_type(rtype, 'seda_' + rtype)
        rename_relation_type(rtype + '_value', 'seda_' + rtype + '_value')
        sync_schema_props_perms('seda_' + rtype + '_value')
    rql('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
        {'cs': cs, 'rt': 'seda_' + rtype})

add_relation_definition('ProfileArchiveObject', 'seda_parent', 'SEDAProfile')
drop_entity_type('ProfileArchive')
sync_schema_props_perms(('ProfileArchiveObject', 'user_cardinality', 'String'))

sync_schema_props_perms(('ProfileArchiveObject', 'seda_parent', 'ProfileArchiveObject'))
sync_schema_props_perms(('ProfileDocument', 'seda_parent', 'ProfileArchiveObject'))

add_relation_type('seda_clone_of')

add_cube('compound')




# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views for OAI-PMH export

Set hierarchy specification
---------------------------

Sets_ are optional construct for selective harvesting. The scheme used to
define the syntax of sets is usually specific to the community using the
OAI-PMH exchange protocol. This OAI-PMH implementation exposes the following
set hierarchy.

- The first level of hierarchy refers to the entity type to perform the
  selective request on, e.g. for a `ListIdentifiers` verb:

      <baseurl>/oai?verb=ListIdentifiers&set=agent

  would return the identifiers of entities of type Agent found in the repository.

- The second level of hierarchy refers to a filtering criterion on selected
  entity type, usually an attribute with respect to the application schema,
  and is tight to a value of this criterion (attribute) to filter entities on.
  For instance:

      <baseurl>/oai?verb=ListIdentifiers&set=agent:role:archival

  would return the identifiers of entities of type Agent with role 'archival'.

.. _Set:
.. _Sets: http://www.openarchives.org/OAI/openarchivesprotocol.html#Set
"""

from datetime import datetime, timedelta

import dateutil.parser
from lxml import etree
from lxml.builder import E, ElementMaker
import pytz

from cubicweb.predicates import ExpectedValuePredicate, match_form_params
from cubicweb.view import View

from cubes.saem_ref import isodate


def utcparse(timestr):
    """Parse a date/time string as an UTC datetime *without* tzinfo."""
    date = dateutil.parser.parse(timestr)
    if date.tzinfo is None:
        if 'T' in timestr:
            raise ValueError('cannot parse a date with time but no timezone')
        else:
            # No time nor tzinfo: done.
            return date
    else:
        # Convert to UTC and drop tzinfo.
        return date.astimezone(pytz.utc).replace(tzinfo=None)


class match_verb(ExpectedValuePredicate):
    """Predicate checking `verb` request form parameter presence and value.

    Return 2 in case of match, for precedence over
    ``match_form_params('verb')``.
    """

    def __call__(self, cls, req, rset=None, **kwargs):
        verb = req.form.get('verb')
        if verb is None:
            return 0
        return 2 * int(verb in self.expected)


def filter_none(mapping):
    """Return a dict from `mapping` with None values filtered out."""
    out = {}
    for key, val in mapping.iteritems():
        if val is not None:
            out[key] = val
    return out


def oai_records(rset):
    """Yield OAIRecord items from a result set."""
    for entity in rset.entities():
        record = entity.cw_adapt_to('IOAIPMHRecord')
        if not record:
            continue
        yield OAIRecord(record)


class OAIError(Exception):
    """Error in OAI-PMH request."""

    def __init__(self, errors):
        super(OAIError, self).__init__()
        self.errors = errors


class OAIRequest(object):
    """Represent an OAI-PMH request."""

    @classmethod
    def from_request(cls, baseurl, request):
        form = request.form
        return cls(
            baseurl,
            setspec=form.get('set'),
            verb=form.get('verb'),
            identifier=form.get('identifier'),
            from_date=form.get('from'),
            until_date=form.get('until'),
            resumption_token=form.get('resumptionToken'),
        )

    def __init__(self, baseurl, verb=None, setspec=None, identifier=None,
                 from_date=None, until_date=None, resumption_token=None):
        self.baseurl = baseurl
        self.verb = verb
        self.setspec = setspec
        self.identifier = identifier
        self.errors = {}
        # Parse "from" and "until" dates, which are, by specification
        # expressed in UTC.
        if from_date is not None:
            from_date = utcparse(from_date)
        if until_date is not None:
            until_date = utcparse(until_date)
        if (from_date is not None and until_date is not None and
                from_date > until_date):
            msg = 'the from argument must be less than or equal to the until argument'
            self.errors['badArgument'] = msg
        self.from_date = from_date
        self.until_date = until_date
        self.resumption_token = resumption_token

    def __repr__(self):
        return etree.tostring(self.to_xml())

    def rset_from_identifier(self, cnx):
        """Return a ResultSet corresponding to request identifier."""
        oai = cnx.vreg['components'].select('oai', cnx)
        return oai.match_identifier(self.identifier)

    def rset_from_setspec(self, cnx):
        """Return a result set from OAI-PMH request set specifier, possibly
        with a resumptionToken information.
        """
        dates = {'from_date': self.from_date,
                 'until_date': self.until_date}
        oai = cnx.vreg['components'].select('oai', cnx)
        try:
            rset, token = oai.match_setspec(
                self.setspec, token=self.resumption_token, **dates)
        except ValueError:
            raise OAIError(
                {'noRecordsMatch': 'invalid set specifier {0}'.format(self.setspec)})
        if not rset:
            if self.resumption_token is not None:
                raise OAIError(
                    {'badResumptionToken': ('The value of the resumptionToken '
                                            'argument is invalid or expired.')})
            raise OAIError(
                {'noRecordsMatch': ('The combination of the values of the '
                                    'from, until, and set arguments results '
                                    'in an empty list.')})
        return rset, token

    def new_token(self, token):
        """Return a resumptionToken XML element or None.

        * return a resumptionToken with a value if there are more result to be
        fetched;
        * return an empty resumptionToken if this response completes and the
        request contains a resumptionToken;
        * None otherwise.
        """
        if token is not None:
            expire = isodate(datetime.now(pytz.utc) + timedelta(hours=1))
            return E.resumptionToken(token, expirationDate=expire)
        elif self.resumption_token is not None:
            return E.resumptionToken()

    def to_xml(self, errors=None):
        if not errors:
            # In cases where the request that generated this response resulted
            # in a badVerb or badArgument error condition, the repository must
            # return the base URL of the protocol request only. Attributes
            # must not be provided in these cases.
            attributes = {
                'verb': self.verb,
                'identifier': self.identifier,
                'set': self.setspec,
            }
            if self.from_date:
                attributes['from'] = isodate(self.from_date)
            if self.until_date:
                attributes['until'] = isodate(self.until_date)
        else:
            attributes = {}
        attributes = filter_none(attributes)
        return E.request(self.baseurl, **attributes)


class OAIResponse(object):
    """Represent an OAI-PMH response."""

    def __init__(self, oai_request):
        self.oai_request = oai_request

    @staticmethod
    def _build_errors(errors):
        """Return a list of <error> tag from `errors` dict."""
        return [E.error(msg, code=code)
                for code, msg in errors.iteritems()]

    def body(self, content=None, errors=None):
        """Return a list of body items of the OAI-PMH response."""
        oai_request = self.oai_request
        if errors:
            return self._build_errors(errors)
        assert content is not None, \
            'unexpected empty content while no error got reported'
        try:
            return [E(oai_request.verb, *content)]
        except OAIError as exc:
            return self._build_errors(exc.errors)

    def to_xml(self, content=None, errors=()):
        date = E.responseDate(isodate())
        request = self.oai_request.to_xml(errors)
        body_elems = self.body(content, errors=errors)
        nsmap = {None: 'http://www.openarchives.org/OAI/2.0/',
                 'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
        maker = ElementMaker(nsmap=nsmap)
        attributes = {
            '{%s}schemaLocation' % nsmap['xsi']: ' '.join([
                'http://www.openarchives.org/OAI/2.0/',
                'http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd'
            ])
        }
        return maker('OAI-PMH', date, request, *body_elems, **attributes)


class OAIRecord(object):
    """Represent an OAI record built from an entity adapted as IOAIPMHRecord.
    """

    def __init__(self, record):
        self.record = record

    def header(self):
        """The <header> part of an OAI-PMH record.

        See http://www.openarchives.org/OAI/openarchivesprotocol.html#header
        for a description of elements of a "header".
        """
        # TODO: add setSpec tag for each record.
        tags = []
        for tag in ('identifier', 'datestamp'):
            value = getattr(self.record, tag)
            if value:
                tags.append(E(tag, value))
        attrs = {}
        if self.record.deleted:
            attrs['status'] = 'deleted'
        return E.header(*tags, **attrs)

    def metadata(self):
        """The <metadata> part of an OAI-PMH record."""
        metadata = self.record.metadata()
        if metadata:
            return E.metadata(etree.XML(metadata))
        return None

    def to_xml(self):
        """Return the <record> XML element."""
        elems = [self.header()]
        metadata = self.metadata()
        if metadata is not None:  # deleted record
            elems.append(metadata)
        return E.record(*elems)


class OAIView(View):
    """Base class for any OAI view, subclasses should either implement
    `errors` or `verb_content` methods.
    """
    __regid__ = 'oai'
    __abstract__ = True
    templatable = False
    content_type = 'text/xml'
    binary = True

    @staticmethod
    def verb_content():
        return

    @staticmethod
    def errors():
        """Return the errors of the OAI-PMH request."""
        return {}

    def __init__(self, *args, **kwargs):
        super(OAIView, self).__init__(*args, **kwargs)
        self.oai_request = OAIRequest.from_request(
            self._cw.build_url('oai'), self._cw)

    def call(self):
        encoding = self._cw.encoding
        assert encoding == 'UTF-8', 'unexpected encoding {0}'.format(encoding)
        self.w('<?xml version="1.0" encoding="%s"?>\n' % encoding)
        oai_response = OAIResponse(self.oai_request)
        # combine errors coming from view selection with those of request
        # processing.
        errors = self.errors()
        errors.update(self.oai_request.errors)
        response_elem = oai_response.to_xml(self.verb_content(), errors=errors)
        self.w(etree.tostring(response_elem, encoding='utf-8'))


class OAIBaseView(OAIView):
    """Base view for OAI-PMH request with no "verb" specified.

    `verb` request parameter is necessary in our implementation.
    """

    def errors(self):
        return {'badVerb': 'no verb specified'}


class OAIWithVerbView(OAIView):
    """Base view for OAI-PMH request with a "verb" specified.

    This view generated an error as the implementation relies on explicit view
    to handle supported verbs.
    """
    __select__ = match_form_params('verb')

    def errors(self):
        """Return the errors of the OAI-PMH request."""
        return {'badVerb': 'illegal verb "{0}"'.format(self.oai_request.verb)}


class OAIListSetsView(OAIView):
    """View handling verb="ListSets" requests."""
    __select__ = match_verb('ListSets')

    @staticmethod
    def build_set(spec, name=u''):
        """Return a "set" element"""
        return E('set', E.setSpec(spec), E.setName(name))

    def verb_content(self):
        oai = self._cw.vreg['components'].select('oai', self._cw)
        for spec, description in oai.setspecs():
            yield self.build_set(spec, name=self._cw._(description))


class OAIListIdentifiersView(OAIView):
    """View handling verb="ListIdentifiers" requests.

    This view returns an error as it handles cases where no "set" selection is
    specified.
    """
    __select__ = match_verb('ListIdentifiers')

    def errors(self):
        return {
            'badArgument': 'ListIdentifiers verb requires "set" restriction'}


class OAIListIdentifiersWithSetView(OAIView):
    """View handling verb="ListIdentifiers" requests with "set" selection."""
    __select__ = match_form_params('set') & match_verb('ListIdentifiers')

    def verb_content(self):
        rset, token = self.oai_request.rset_from_setspec(self._cw)
        for record in oai_records(rset):
            yield record.header()
        new_token = self.oai_request.new_token(token)
        if new_token is not None:
            yield new_token


class OAIListRecordsView(OAIView):
    """View handling verb="ListRecords" requests.

    This view returns an error as it handles cases where no "set" selection is
    specified.
    """
    __select__ = match_verb('ListRecords')

    def errors(self):
        return {
            'badArgument': 'ListRecords verb requires "set" restriction'}


class OAIListRecordsWithSetView(OAIView):
    """View handling verb="ListRecords" requests with "set" selection."""
    __select__ = match_form_params('set') & match_verb('ListRecords')
    # TODO: handle metadataPrefix, which is a required argument of this verb.

    def verb_content(self):
        rset, token = self.oai_request.rset_from_setspec(self._cw)
        for record in oai_records(rset):
            yield record.to_xml()
        new_token = self.oai_request.new_token(token)
        if new_token is not None:
            yield new_token


class OAIGetRecordErrorView(OAIView):
    """View handling verb="GetRecord" requests with errors.

    This view returns an error as it handles cases where no "identifier" selection is
    specified.
    """
    __select__ = match_verb('GetRecord')

    def errors(self):
        return {'badArgument': ('GetRecord verb requires "identifier" and '
                                '"metadataPrefix" arguments')}


class OAIGetRecordView(OAIView):
    """View handling verb="GetRecord" with proper arguments."""
    __select__ = match_form_params('identifier') & match_verb('GetRecord')
    # TODO: handle metadataPrefix, which is a required argument of this verb.

    def verb_content(self):
        rset = self.oai_request.rset_from_identifier(self._cw)
        for record in oai_records(rset):
            if record is not None:
                yield record.to_xml()
                break
        else:
            msg = 'no entity with OAI identifier {0} in repository'.format(
                self.oai_request.identifier)
            raise OAIError({'idDoesNotExist': msg})



sync_schema_props_perms('contact_point')
sync_schema_props_perms('cwuri')
sync_schema_props_perms('ark')

drop_attribute('ChronologicalRelation', 'start_date')
drop_attribute('ChronologicalRelation', 'end_date')

add_cube('vtimeline')

for eid in set(x for x, in rql('Any S WHERE X is ConceptScheme, X cw_source S, S name != "system"', build_descr=False)):
    source = cnx.entity_from_eid(eid)
    create_entity('SKOSSource', name=source.name,
                  url=source.url, through_cw_source=source)
    commit()

add_entity_type('SEDAFileTypeCode')
add_relation_type('seda_file_type_code')
add_relation_type('seda_file_type_code_value')

add_entity_type('SEDACharacterSetCode')
add_relation_type('seda_character_set_code')
add_relation_type('seda_character_set_code_value')

add_entity_type('SEDADocumentTypeCode')
add_relation_type('seda_document_type_code')
add_relation_type('seda_document_type_code_value')

# Add new SEDA schemes by calling back create_seda_schemes script (no-op for
# existing schemes).
from os.path import join, dirname
process_script(join(dirname(__file__), 'create_seda_schemes.py'))



# coding: utf-8
for rtype in ('seda_file_type_code', 'seda_document_type_code', 'seda_character_set_code'):
    sync_schema_props_perms(rtype)
    sync_schema_props_perms(rtype + '_value')


add_attribute('SEDADocumentTypeCode', 'user_annotation')

for title in (u'SEDA : Formats de fichier source',
              u'SEDA : jeu de caractères de codage',
              u'SEDA : codes des types de contenu'):
    find('ConceptScheme', title=title).one().cw_delete()
    commit()

from os.path import join, dirname
process_script(join(dirname(__file__), 'create_seda_schemes.py'))


# add word table
sqls = """
CREATE EXTENSION  pg_trgm;;
CREATE TABLE words (
  etype VARCHAR(64) NOT NULL,
  word VARCHAR(100) NOT NULL
);;
CREATE INDEX words_unique_idx ON words (etype, word);;
CREATE INDEX words_word_idx ON words USING gin(word gin_trgm_ops)
"""
for statement in sqls.split(';;'):
    sql(statement)

print 'you should now rebuild the text index using `cubicweb-ctl db-rebuild-fti %s`' % config.appid



from urlparse import urlparse, urlunparse

add_relation_type('scheme_relation')
for etype in ('SEDAProfile', 'ProfileArchive', 'ProfileDocument', 'ProfileArchiveObject',
              'SEDAAppraisalRule', 'SEDAAppraisalRuleDuration', 'SEDAAppraisalRuleCode',
              'SEDAContentDescription', 'SEDAAccessRestrictionCode', 'SEDAKeyword',
              'SEDAName', 'SEDADescription', 'SEDADescriptionLevel', 'SEDADate'):
    add_entity_type(etype)

from os.path import join, dirname
process_script(join(dirname(__file__), 'create_seda_schemes.py'))

sync_schema_props_perms('MirrorEntity')
for agent_mirror in rql('Any X,ID WHERE X extid ID, X mirror_of A, A is Agent').entities():
    agent_mirror.cw_set(extid='Agent ' + agent_mirror.extid)
commit()

add_attribute('AsalaeConnector', 'url')
add_attribute('AsalaeConnector', 'password')
sync_schema_props_perms('AsalaeConnector')
add_attribute('AlfrescoConnector', 'sedaprofiles_node')

for alf_connector in find_entities('AlfrescoConnector'):
    if alf_connector.port != 80:
        parsed = list(urlparse(alf_connector.url))
        parsed[1] += ':%d' % alf_connector.port
        url = urlunparse(parsed)
        alf_connector.cw_set(url=url)
        commit()
drop_attribute('AlfrescoConnector', 'port')



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for Agent and its associated classes"""

from cubicweb.predicates import is_instance
from cubicweb.entities import AnyEntity, fetch_config

from cubes.skos import rdfio
from cubes.skos.entities import AbstractRDFAdapter

from cubes.saem_ref.entities import oai


def _register_agent_prov_mapping(reg):  # XXX move to the prov cube
    """Register RDF mapping for PROV-O entity types related to Agent.
    """
    reg.register_prefix('prov', 'http://www.w3.org/ns/prov#')
    # reg.register_etype_equivalence('Agent', 'prov:Agent')
    reg.register_etype_equivalence('Activity', 'prov:Activity')
    reg.register_attribute_equivalence('Activity', 'type', 'prov:type')
    reg.register_attribute_equivalence('Activity', 'description', 'prov:label')
    reg.register_attribute_equivalence('Activity', 'start', 'prov:startedAtTime')
    reg.register_attribute_equivalence('Activity', 'end', 'prov:endedAtTime')
    reg.register_relation_equivalence('Activity', 'associated_with', 'Agent',
                                      'prov:wasAssociatedWith')
    reg.register_relation_equivalence('Activity', 'generated', 'Agent', 'prov:generated')
    reg.register_relation_equivalence('Activity', 'used', 'Agent', 'prov:used')


class Agent(AnyEntity):
    __regid__ = 'Agent'
    fetch_attrs, cw_fetch_order = fetch_config(('name', 'agent_kind'))

    @property
    def kind(self):
        """The kind of agent"""
        return self.agent_kind[0].name

    def has_role(self, role):
        """Return true if the agent has the archival role `role` else False"""
        return any(archival_role for archival_role in self.archival_role
                   if archival_role.name == role)

    @property
    def printable_kind(self):
        """The kind of agent, for display"""
        return self.agent_kind[0].printable_value('name')


def _fill_agent_rdf_mapping(reg, agent):
    reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
    reg.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
    foaf_kind = {'person': 'Person',
                 'authority': 'Organization',
                 'family': 'Group'}.get(agent.kind, 'Agent')
    reg.register_etype_equivalence('Agent', 'foaf:' + foaf_kind)
    # Standard metadata.
    reg.register_attribute_equivalence('Agent', 'name', 'foaf:name')
    # Additionnal metadata for Asalae export.
    reg.register_attribute_equivalence('Agent', 'ark', 'dc:identifier')


def _complete_agent_rdf_mapping(reg):
    reg.register_prefix('dct', 'http://purl.org/dc/terms/')
    reg.register_prefix('schema_org', 'http://schema.org/')
    reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')
    reg.register_attribute_equivalence('Agent', 'creation_date', 'dct:created')
    reg.register_attribute_equivalence('Agent', 'modification_date', 'dct:modified')
    reg.register_attribute_equivalence('Agent', 'start_date', 'schema_org:startDate')
    reg.register_attribute_equivalence('Agent', 'end_date', 'schema_org:endDate')
    reg.register_etype_equivalence('PostalAddress', 'vcard:Location')
    reg.register_attribute_equivalence('PostalAddress', 'street', 'vcard:street-address')
    reg.register_attribute_equivalence('PostalAddress', 'postalcode', 'vcard:postal-code')
    reg.register_attribute_equivalence('PostalAddress', 'city', 'vcard:locality')
    reg.register_attribute_equivalence('PostalAddress', 'country', 'vcard:country-name')
    reg.register_attribute_equivalence('PostalAddress', 'state', 'vcard:region')


class AgentRDFListAdapter(AbstractRDFAdapter):
    __regid__ = 'RDFList'
    __select__ = is_instance('Agent')

    def register_rdf_mapping(self, reg):
        _fill_agent_rdf_mapping(reg, self.entity)


class AgentRDFPrimaryAdapter(AgentRDFListAdapter):
    __regid__ = 'RDFPrimary'

    def register_rdf_mapping(self, reg):
        """RDF mapping for Agent entity type."""
        super(AgentRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _complete_agent_rdf_mapping(reg)
        reg.register_prefix('saem', 'http://www.logilab.org/saem/0#')
        reg.register_relation_equivalence('Agent', 'contact_point', 'Agent',
                                          'schema_org:contactPoint')
        reg.register_relation_equivalence('Agent', 'archival_agent', 'Agent',
                                          'saem:archivalAgent')

    def fill(self, graph):
        super(AgentRDFPrimaryAdapter, self).fill(graph)
        reg = self.registry
        generator = rdfio.RDFGraphGenerator(graph)
        # Export archival roles for the agent
        for archival_role in self.entity.archival_role:
            graph.add(graph.uri(self.entity.absolute_url()),
                      graph.uri(reg.normalize_uri('vcard:role')),
                      archival_role.name)
        # Export contact agent
        if self.entity.contact_point:
            contact_agent = self.entity.contact_point[0]
            # this is necessary to generate proper foaf:type for the contact agent, else it will
            # reuse the foaf:type of the exported agent
            contact_reg = rdfio.RDFRegistry()
            _fill_agent_rdf_mapping(contact_reg, contact_agent)
            _complete_agent_rdf_mapping(contact_reg)
            generator.add_entity(contact_agent, contact_reg)
        # Export addresses for the agent
        for place in self.entity.reverse_place_agent:
            for address in place.place_address:
                graph.add(graph.uri(self.entity.absolute_url()),
                          graph.uri(reg.normalize_uri('vcard:hasAddress')),
                          graph.uri(address.absolute_url()))
                if place.role:
                    graph.add(graph.uri(address.absolute_url()),
                              graph.uri(reg.normalize_uri('vcard:role')),
                              place.role)
                generator.add_entity(address, reg)
        # Export chronological relations
        for rtype_name, reverse_rtype_name, obj_property, subj_property in [
            ('reverse_chronological_successor', 'chronological_predecessor', 'dct:isReplacedBy',
             'dct:replaces'),
            ('reverse_chronological_predecessor', 'chronological_successor', 'dct:replaces',
             'dct:isreplacedBy')
        ]:
            for chrono_relation in getattr(self.entity, rtype_name):
                for related_agent in getattr(chrono_relation, reverse_rtype_name):
                    graph.add(graph.uri(related_agent.absolute_url()),
                              graph.uri(reg.normalize_uri(obj_property)),
                              graph.uri(self.entity.absolute_url()))
                    graph.add(graph.uri(self.entity.absolute_url()),
                              graph.uri(reg.normalize_uri(subj_property)),
                              graph.uri(related_agent.absolute_url()))


class AgentOAIPMHRecordAdapter(oai.OAIPMHActiveRecordAdapter):
    """OAI-PMH adapter for Agent entity type."""
    __select__ = oai.OAIPMHActiveRecordAdapter.__select__ & is_instance('Agent')
    metadata_view = 'primary.rdf'

    @classmethod
    def set_definition(cls):
        specifier = oai.PublicETypeOAISetSpec(
            'Agent', identifier_attribute=cls.identifier_attribute)
        specifier['role'] = oai.RelatedEntityOAISetSpec(
            'archival_role', 'ArchivalRole', 'name',
            description=u'An agent with {0} archival role')
        specifier['kind'] = oai.RelatedEntityOAISetSpec(
            'agent_kind', 'AgentKind', 'name',
            description=u'An agent of kind {0}',
            exclude=['unknown-agent-kind'])
        return specifier


class AgentKind(AnyEntity):
    __regid__ = 'AgentKind'
    fetch_attrs, cw_fetch_order = fetch_config(('name',))


class ArchivalRole(AnyEntity):
    __regid__ = 'AgentKind'
    fetch_attrs, cw_fetch_order = fetch_config(('name',))


class ChronologicalRelation(AnyEntity):
    __regid__ = 'ChronologicalRelation'

    def dc_description(self):
        if self.description:
            return self.description


class EACResourceRelation(AnyEntity):
    __regid__ = 'EACResourceRelation'

    @property
    def agent(self):
        return self.resource_relation_agent[0]

    @property
    def resource(self):
        return self.resource_relation_resource[0]

    def dc_title(self):
        agent_title = self.agent.dc_title()
        if self.agent_role:
            agent_title += u' (%s)' % self.printable_value('agent_role')
        resource_title = self.resource.dc_title()
        if self.resource_role:
            resource_title += u' (%s)' % self.printable_value('resource_role')
        return (self._cw._('Relation from %(from)s to %(to)s ') %
                {'from': agent_title,
                 'to': resource_title})


class Citation(AnyEntity):
    __regid__ = 'Citation'

    def dc_title(self):
        title = []
        if self.note:
            title.append(self.note)
        if self.uri:
            title.append(u'<{0}>'.format(self.uri))
        if title:
            return u' '.join(title)
        return u'{0} #{1}'.format(self._cw._('Citation'), self.eid)


class SameAsMixIn(object):
    """Mix-in class for entity types supporting vocabulary_source and equivalent_concept relations
    """

    @property
    def scheme(self):
        return self.vocabulary_source and self.vocabulary_source[0] or None

    @property
    def concept(self):
        return self.equivalent_concept and self.equivalent_concept[0] or None


class AgentPlace(SameAsMixIn, AnyEntity):
    __regid__ = 'AgentPlace'


class AgentFunction(SameAsMixIn, AnyEntity):
    __regid__ = 'AgentFunction'


class Mandate(SameAsMixIn, AnyEntity):
    __regid__ = 'Mandate'


class LegalStatus(SameAsMixIn, AnyEntity):
    __regid__ = 'LegalStatus'


class Occupation(SameAsMixIn, AnyEntity):
    __regid__ = 'Occupation'



# coding: utf-8
"""Create concept schemes used by SEDA profile data model"""

from os.path import join, dirname

from cubicweb.dataimport.importer import HTMLImportLog

from cubes.skos.sobjects import lcsv_extentities, store_skos_extentities

from cubes.saem_ref.sobjects.skos import _store


def lcsv_import(cnx, store, fname, scheme_uri):
    import_log = HTMLImportLog(fname)
    stream = open(join(dirname(__file__), 'data', fname))
    entities = lcsv_extentities(stream, scheme_uri, '\t', 'utf-8')
    store_skos_extentities(cnx, store, entities, import_log,
                           raise_on_error=True, extid_as_cwuri=False)


def import_seda_schemes(cnx):
    store = _store(cnx)
    for title, rtype, comment, fname in (
            (u"SEDA : Codes de restriction d'accès",
             'seda_access_restriction_code',
             'access control',
             'seda_access_control.csv'),
            (u'SEDA : Sort final',
             'seda_appraisal_rule_code',
             'appraisal rule',
             'seda_appraisal_rule_code.csv'),
            (u'SEDA : Niveaux de description de contenu',
             'seda_description_level',
             'description level',
             'seda_description_level.csv'),
            (u'SEDA : Formats de fichier source',
             'seda_file_type_code',
             'file type code',
             'seda_file_type_code.csv'),
            (u'SEDA : Jeu de caractères de codage',
             'seda_character_set_code',
             'character set code',
             'seda_character_set_code.csv'),
            (u'SEDA : Codes des types de contenu',
             'seda_document_type_code',
             'document type code',
             'seda_document_type_code.csv'),
    ):
        if not cnx.find('ConceptScheme', title=title):
            scheme = cnx.create_entity('ConceptScheme', title=title)
            inserted = cnx.execute(
                'SET X scheme_relation CR WHERE CR name "%s", X eid %%(x)s' % rtype,
                {'x': scheme.eid})
            assert inserted, comment
            lcsv_import(cnx, store, fname, scheme.cwuri)
    store.flush()
    store.commit()
    store.finish()


print('-> creating SEDA concept schemes')
import_seda_schemes(cnx)

if repo.system_source.dbdriver == 'postgres':
    # massive store is used, which doesn't handle full-text indexation
    reindex_entities(['ConceptScheme', 'Concept'])



# coding: utf-8
"""cubicweb-saem_ref unit tests for dataimport"""

import datetime
import os.path
import sys
from itertools import count, imap

from logilab.common.testlib import TestCase

from cubicweb.dataimport.importer import ExtEntity, SimpleImportLog
from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref.dataimport import eac


def mock_(string):
    return string


class EACXMLParserTC(TestCase):

    def test_parse_FRAD033_EAC_00001(self):
        expected = [
            ('Agent',
             'FRAD033_EAC_00001',
             {'isni': set([u'22330001300016']),
              'name': set([u'Gironde. Conseil général']),
              'start_date': set([datetime.date(1800, 1, 1)]),
              'end_date': set([datetime.date(2099, 1, 1)]),
              'agent_kind': set(['agentkind/authority']),
             },
            ),
            ('EACSource',
             '0',
             {'source_agent': set(['FRAD033_EAC_00001']),
              'title': set([u'1. Ouvrages imprimés...']),
              'description': set([u'des bouquins']),
              'description_format': set([u'text/plain']),
             },
            ),
            ('EACSource',
             '1',
             {'source_agent': set(['FRAD033_EAC_00001']),
              'url': set([u'http://archives.gironde.fr']),
              'title': set([u'Site des Archives départementales de la Gironde']),
             },
            ),
            ('Activity',
             '2',
             {'type': set([u'create']),
              'generated': set(['FRAD033_EAC_00001']),
              'start': set([datetime.datetime(2013, 4, 24, 5, 34, 41)]),
              'end': set([datetime.datetime(2013, 4, 24, 5, 34, 41)]),
              'description': set([u'bla bla']),
              'description_format': set([u'text/plain']),
             },
            ),
            ('AgentKind',
             'agentkind/person',
             {'name': set([u'person'])},
            ),
            ('Agent',
             '3',
             {'name': set([u'Delphine Jamet']),
              'agent_kind': set(['agentkind/person']),
             },
            ),
            ('Activity',
             '4',
             {'associated_with': set(['3']),
              'generated': set(['FRAD033_EAC_00001']),
              'type': set([u'modify']),
              'start': set([datetime.datetime(2015, 1, 15, 7, 16, 33)]),
              'end': set([datetime.datetime(2015, 1, 15, 7, 16, 33)]),
             },
            ),
            ('AgentKind',
             'agentkind/authority',
             {'name': set([u'authority'])},
            ),
            ('PostalAddress',
             '5',
             {'street': set([u'1 Esplanade Charles de Gaulle']),
              'postalcode': set([u'33074']),
              'city': set([u' Bordeaux Cedex']),
             },
            ),
            ('AgentPlace',
             '6',
             {'name': set([u'Bordeaux (Gironde, France)']),
              'role': set([u'siege']),
              'place_agent': set(['FRAD033_EAC_00001']),
              'place_address': set(['5']),
              'equivalent_concept': set([u'http://catalogue.bnf.fr/ark:/12148/cb152418385']),
             },
            ),
            ('AgentPlace',
             '7',
             {'name': set([u'Toulouse (France)']),
              'place_agent': set(['FRAD033_EAC_00001']),
              'role': set([u'domicile']),
             },
            ),
            ('AgentPlace',
             '8',
             {'name': set([u'Lit']),
              'place_agent': set(['FRAD033_EAC_00001']),
              'role': set([u'dodo']),
             },
            ),
            ('LegalStatus',
             '9',
             {'term': set([u'Collectivité territoriale']),
              'start_date': set([datetime.date(1234, 1, 1)]),
              'end_date': set([datetime.date(3000, 1, 1)]),
              'description': set([u'Description du statut']),
              'description_format': set([u'text/plain']),
              'legal_status_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('Mandate',
             '10',
             {'term': set([u'1. Constitutions françaises']),
              'description': set([u'Description du mandat']),
              'description_format': set([u'text/plain']),
              'mandate_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('History',
             '11',
             {'text': set([u'<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">La loi du 22 décembre 1789, en divisant ...</p>']),  # noqa
              'text_format': set([u'text/html']),
              'history_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('Structure',
             '12',
             {'description': set([u'<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">Pour accomplir ses missions ...</p>']),  # noqa
              'description_format': set([u'text/html']),
              'structure_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('AgentFunction',
             '13',
             {'description': set([u'<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">Quatre grands domaines de compétence...</p>']),
              'description_format': set([u'text/html']),
              'function_agent': set(['FRAD033_EAC_00001']),
             },
            ),
            ('AgentFunction',
             '14',
             {'name': set([u'action sociale']),
              'function_agent': set(['FRAD033_EAC_00001']),
              'description': set([u'''<p xmlns="urn:isbn:1-931666-33-4" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xlink="http://www.w3.org/1999/xlink">1. Solidarité
            blablabla.</p>''']),
              'description_format': set([u'text/html']),
              'equivalent_concept': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200']),
             },
            ),
            ('AgentFunction',
             '15',
             {'name': set([u'environnement']),
              'function_agent': set(['FRAD033_EAC_00001']),
              'equivalent_concept': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074']),
             },
            ),
            ('Citation',
             '16',
             {'note': set([u'la bible']),
             },
            ),
            ('Occupation',
             '17',
             {'term': set([u'Réunioniste']),
              'start_date': set([datetime.date(1987, 1, 1)]),
              'end_date': set([datetime.date(2099, 1, 1)]),
              'description': set([u'Organisation des réunions ...']),
              'description_format': set([u'text/plain']),
              'occupation_agent': set(['FRAD033_EAC_00001']),
              'has_citation': set(['16']),
              'equivalent_concept': set(['http://pifgadget.com']),
             },
            ),
            ('Agent',
             'CG33-DIRADSJ',
             {'name': set([u"Gironde. Conseil général. Direction de l'administration et de la "
                           u"sécurité juridique"]),
              'agent_kind': set(['agentkind/unknown-agent-kind']),
             },
            ),
            ('HierarchicalRelation',
             '18',
             {'start_date': set([datetime.date(2008, 1, 1)]),
              'end_date': set([datetime.date(2099, 1, 1)]),
              'description': set([u'Coucou']),
              'description_format': set([u'text/plain']),
              'hierarchical_parent': set(['CG33-DIRADSJ']),
              'hierarchical_child': set(['FRAD033_EAC_00001']),
             },
            ),
            ('Agent',
             'whatever',
             {'name': set([u'CG32']),
              'agent_kind': set(['agentkind/unknown-agent-kind']),
             },
            ),
            ('ChronologicalRelation',
             '19',
             {'chronological_predecessor': set(['whatever']),
              'chronological_successor': set(['FRAD033_EAC_00001']),
             },
            ),
            ('AssociationRelation',
             '20',
             {'association_from': set(['FRAD033_EAC_00001']),
              'association_to': set(['agent-x']),
             },
            ),
            ('EACResourceRelation',
             '21',
             {'agent_role': set([u'creatorOf']),
              'resource_role': set([u'Fonds d\'archives']),
              'resource_relation_resource': set([
                  'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N']),
              'resource_relation_agent': set(['FRAD033_EAC_00001']),
              'start_date': set([datetime.date(1673, 1, 1)]),
              'end_date': set([datetime.date(1963, 1, 1)]),
             },
            ),
            ('ExternalUri',
             'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N',
             {'uri': set([u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N']),
              'cwuri': set([u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N'])},
            ),
            ('ExternalUri',
             'agent-x',
             {'uri': set([u'agent-x']), 'cwuri': set([u'agent-x'])},
            ),
            ('ExternalUri',
             'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200',
             {'uri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200']),
              'cwuri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200'])},
            ),
            ('ExternalUri',
             'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074',
             {'uri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074']),
              'cwuri': set([u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074'])},
            ),
            ('ExternalUri',
             'http://catalogue.bnf.fr/ark:/12148/cb152418385',
             {'uri': set([u'http://catalogue.bnf.fr/ark:/12148/cb152418385']),
              'cwuri': set([u'http://catalogue.bnf.fr/ark:/12148/cb152418385'])},
            ),
            ('ExternalUri',
             'http://pifgadget.com',
             {'uri': set([u'http://pifgadget.com']),
              'cwuri': set([u'http://pifgadget.com'])},
            ),
        ]
        expected = [ExtEntity(*vals) for vals in expected]
        fpath = self.datapath('EAC/FRAD033_EAC_00001_simplified.xml')
        import_log = SimpleImportLog(fpath)
        # Use a predictable extid_generator.
        extid_generator = imap(str, count()).next
        importer = eac.EACCPFImporter(fpath, import_log, mock_,
                                      extid_generator=extid_generator)
        entities = list(importer.external_entities())
        self.check_external_entities(entities, expected)
        visited = set([])
        for x in importer._visited.values():
            visited.update(x)
        self.assertItemsEqual(visited, [x.extid for x in expected])
        # Gather not-visited tag by name and group source lines.
        not_visited = {}
        for tagname, sourceline in importer.not_visited():
            not_visited.setdefault(tagname, set([])).add(sourceline)
        self.assertEqual(not_visited,
                         {'maintenanceStatus': set([10]),
                          'publicationStatus': set([12]),
                          'maintenanceAgency': set([14]),
                          'languageDeclaration': set([19]),
                          'conventionDeclaration': set([23, 33, 42]),
                          'localControl': set([52]),
                          'generalContext': set([188]),
                          'nameEntry': set([112]),  # alternative form.
                          'source': set([74]),  # empty.
                          'biogHist': set([193]),  # empty.
                          'structureOrGenealogy': set([186]),  # empty.
                          })

    def test_mandate_under_mandates(self):
        """In FRAD033_EAC_00003.xml, <mandate> element are within <mandates>."""
        fpath = self.datapath('EAC/FRAD033_EAC_00003.xml')
        import_log = SimpleImportLog(fpath)
        # Use a predictable extid_generator.
        extid_generator = imap(str, count()).next
        importer = eac.EACCPFImporter(fpath, import_log, mock_,
                                      extid_generator=extid_generator)
        entities = list(importer.external_entities())
        expected_terms = [
            u'Code du patrimoine, Livre II',
            u'Loi du 5 brumaire an V [26 octobre 1796]',
            (u'Loi du 3 janvier 1979 sur les archives, accompagnée de ses décrets\n'
             u'                        d’application datant du 3 décembre.'),
            u'Loi sur les archives du 15 juillet 2008',
        ]
        self.assertCountEqual([next(iter(x.values['term'])) for x in entities
                               if x.etype == 'Mandate' and 'term' in x.values],
                              expected_terms)
        mandate_with_link = next(x for x in entities if x.etype == 'Mandate' and
                                 u'Code du patrimoine, Livre II' in x.values['term'])
        extid = next(iter(mandate_with_link.values['has_citation']))
        url = u'http://www.legifrance.gouv.fr/affichCode.do?idArticle=LEGIARTI000019202816'
        citation = next(x for x in entities if x.etype == 'Citation'
                        and url in x.values['uri'])
        self.assertEqual(extid, citation.extid)

    def test_agentfunction_within_functions_tag(self):
        """In FRAD033_EAC_00003.xml, <function> element are within <functions>
        not <description>.
        """
        fpath = self.datapath('EAC/FRAD033_EAC_00003.xml')
        import_log = SimpleImportLog(fpath)
        # Use a predictable extid_generator.
        extid_generator = imap(str, count()).next
        importer = eac.EACCPFImporter(fpath, import_log, mock_,
                                      extid_generator=extid_generator)
        entities = importer.external_entities()
        self.assertCountEqual(
            [x.values['name'].pop() for x in entities
             if x.etype == 'AgentFunction' and 'name' in x.values],
            [u'contr\xf4le', u'collecte', u'classement', u'restauration', u'promotion'])

    def ctx_assert(self, method, actual, expected, ctx, msg=None):
        """Wrap assertion method with a context message"""
        try:
            getattr(self, method)(actual, expected, msg=msg)
        except AssertionError as exc:
            msg = str(exc)
            if ctx:
                msg = ('[%s] ' % ctx) + msg
            raise AssertionError(msg), None, sys.exc_info()[-1]

    def check_external_entities(self, entities, expected):
        entities = extentities2dict(entities)
        expected = extentities2dict(expected)
        etypes, expected_etypes = entities.keys(), expected.keys()
        self.ctx_assert('assertCountEqual', etypes, expected_etypes, ctx='etypes')
        for etype, edict in expected.iteritems():
            entities_etype = entities[etype]
            extids, expected_extids = entities_etype.keys(), edict.keys()
            self.ctx_assert('assertCountEqual', extids, expected_extids,
                            ctx='%s/extids' % etype)
            for extid, values in edict.iteritems():
                self.ctx_assert('assertEqual', entities_etype[extid], values,
                                ctx='%s/%s/values' % (etype, extid))


def extentities2dict(entities):
    edict = {}
    for extentity in entities:
        edict.setdefault(extentity.etype, {})[extentity.extid] = extentity.values
    return edict


def eac_import(cnx, fpath):
    import_log = SimpleImportLog(os.path.basename(fpath))
    created, updated, _ = cnx.call_service(
        'saem_ref.eac-import', stream=fpath, import_log=import_log,
        raise_on_error=True)
    return created, updated


class EACDataImportTC(CubicWebTC):

    def test_FRAD033_EAC_00001(self):
        fpath = self.datapath('EAC/FRAD033_EAC_00001_simplified.xml')
        with self.admin_access.repo_cnx() as cnx:
            # create a skos concept to ensure it's used instead of a ExternalUri
            scheme = cnx.create_entity('ConceptScheme')
            scheme.add_concept(u'environnement',
                               cwuri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074')
            cnx.commit()
            created, updated = eac_import(cnx, fpath)
            self.assertEqual(len(created), 30)
            person_eid = cnx.find('AgentKind', name=u'person')[0][0]
            authority_eid = cnx.find('AgentKind', name=u'authority')[0][0]
            self.assertEqual(updated, set([authority_eid, person_eid]))
            rset = cnx.find('Agent', isni=u'22330001300016')
            self.assertEqual(len(rset), 1)
            agent = rset.one()
            self.assertEqual(agent.kind, 'authority')
            self.assertEqual(agent.start_date, datetime.date(1800, 1, 1))
            self.assertEqual(agent.end_date, datetime.date(2099, 1, 1))
            address = agent.postal_address[0]
            self.assertEqual(address.street, u'1 Esplanade Charles de Gaulle')
            self.assertEqual(address.postalcode, u'33074')
            self.assertEqual(address.city, u' Bordeaux Cedex')
            rset = cnx.execute('Any R,N WHERE P place_agent A, A eid %(eid)s, P role R, P name N',
                               {'eid': agent.eid})
            self.assertCountEqual(rset.rows,
                                  [[u'siege', u'Bordeaux (Gironde, France)'],
                                   [u'domicile', u'Toulouse (France)'],
                                   [u'dodo', u'Lit']])
            self.assertEqual(len(agent.reverse_function_agent), 3)
            self.set_description('check Structure')
            self._check_structure(cnx, agent)
            self.set_description('check History')
            self._check_history(cnx, agent)
            self.set_description('check Mandate')
            self._check_mandate(cnx, agent)
            self.set_description('check Occupation')
            self._check_occupation(cnx, agent)
            self.set_description('check LegalStatus')
            self._check_legal_status(cnx, agent)
            self.set_description('check relations')
            self._check_eac_relations(cnx, agent)
            self.set_description('check vocabulary source')
            self._check_equivalent_concept(cnx, agent)
            self.set_description('check control')
            self._check_control(cnx, agent)

    def _check_structure(self, cnx, agent):
        rset = cnx.find('Structure', structure_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('description',
                                                    format=u'text/plain').strip(),
                         u'Pour accomplir ses missions ...')

    def _check_history(self, cnx, agent):
        rset = cnx.find('History', history_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('text',
                                                    format=u'text/plain').strip(),
                         u'La loi du 22 décembre 1789, en divisant ...')

    def _check_mandate(self, cnx, agent):
        rset = cnx.find('Mandate', mandate_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('description',
                                                    format=u'text/plain').strip(),
                         u'Description du mandat')

    def _check_occupation(self, cnx, agent):
        occupation = cnx.find('Occupation', occupation_agent=agent).one()
        self.assertEqual(occupation.term, u'Réunioniste')
        citation = occupation.has_citation[0]
        self.assertEqual(citation.note, u'la bible')
        voc = occupation.equivalent_concept[0]
        self.assertEqual(voc.uri, u'http://pifgadget.com')

    def _check_legal_status(self, cnx, agent):
        rset = cnx.find('LegalStatus', legal_status_agent=agent)
        self.assertEqual(len(rset), 1)
        self.assertEqual(rset.one().printable_value('description',
                                                    format=u'text/plain').strip(),
                         u'Description du statut')

    def _check_eac_relations(self, cnx, agent):
        relation = cnx.find('HierarchicalRelation').one()
        self.assertEqual(relation.printable_value('description',
                                                  format='text/plain'),
                         u'Coucou')
        other_agent = cnx.find('Agent', name=u"Gironde. Conseil général. Direction de "
                               u"l'administration et de la sécurité juridique").one()
        self.assertEqual(relation.hierarchical_parent[0], other_agent)
        relation = cnx.find('AssociationRelation').one()
        self.assertEqual(relation.association_from[0], agent)
        other_agent = cnx.find('ExternalUri', uri=u'agent-x').one()
        self.assertEqual(other_agent.cwuri, 'agent-x')
        self.assertEqual(relation.association_to[0], other_agent)
        rset = cnx.find('EACResourceRelation', agent_role=u'creatorOf')
        self.assertEqual(len(rset), 1)
        rrelation = rset.one()
        self.assertEqual(rrelation.resource_relation_agent[0], agent)
        exturi = rrelation.resource_relation_resource[0]
        self.assertEqual(exturi.uri,
                         u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N')

    def _check_equivalent_concept(self, cnx, agent):
        functions = dict((f.name, f) for f in agent.reverse_function_agent)
        self.assertEqual(functions['action sociale'].equivalent_concept[0].cwuri,
                         'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200')
        self.assertEqual(functions['action sociale'].equivalent_concept[0].cw_etype,
                         'ExternalUri')
        self.assertEqual(functions['environnement'].equivalent_concept[0].cwuri,
                         'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074')
        self.assertEqual(functions['environnement'].equivalent_concept[0].cw_etype,
                         'Concept')
        self.assertEqual(functions['environnement'].vocabulary_source[0].eid,
                         functions['environnement'].equivalent_concept[0].scheme.eid)
        place = cnx.find('AgentPlace', role=u'siege').one()
        self.assertEqual(place.equivalent_concept[0].cwuri,
                         'http://catalogue.bnf.fr/ark:/12148/cb152418385')

    def _check_control(self, cnx, agent):
        rset = cnx.find('EACSource')
        self.assertEqual(len(rset), 2)
        rset = cnx.execute('Any A WHERE A generated X, X eid %s' % agent.eid)
        # Two activities imported from EAC-CPF file, one created by our hook.
        self.assertEqual(len(rset), 3)
        rset = cnx.execute('Any A WHERE A associated_with X, X name "Delphine Jamet"')
        self.assertEqual(len(rset), 1)

    def test_multiple_imports(self):
        with self.admin_access.repo_cnx() as cnx:
            for fname in ('FRAD033_EAC_00001.xml', 'FRAD033_EAC_00003.xml',
                          'FRAD033_EAC_00071.xml'):
                self.set_description('importing ' + fname)
                fpath = self.datapath('EAC/' + fname)
                created, updated = eac_import(cnx, fpath)
            self.set_description('check')
            # Imported agent do not have a CWUSer associated.
            rset = cnx.execute('Any X WHERE X is Agent, NOT EXISTS(X agent_user U)')
            self.assertEqual(len(rset), 15)


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



add_attribute('SEDAAppraisalRule', 'user_cardinality')

add_entity_type('AssociationRelation')
add_entity_type('ChronologicalRelation')
add_entity_type('HierarchicalRelation')

for entity in find_entities('AgentRelation'):
    etype, rtype_from, rtype_to = {
        'association': ('AssociationRelation', 'association_from',
                        'association_to'),
        'chronological': ('ChronologicalRelation', 'chronological_predecessor',
                          'chronological_successor'),
        'hierarchical': ('HierarchicalRelation', 'hierarchical_parent',
                         'hierarchical_child'),
    }[entity.relationship]
    kwargs = {rtype_from: entity.agent_relationship_from,
              rtype_to: entity.agent_relationship_to}
    create_entity(etype, description=entity.description,
                  description_format=entity.description_format,
                  **kwargs)

drop_entity_type('AgentRelation')

add_relation_type('contact_point')

drop_attribute('SEDAKeyword', 'value')
add_relation_definition('SEDAKeyword', 'seda_keyword_value', 'Concept')
rql('DELETE SEDAKeyword K WHERE NOT K seda_keyword_scheme S')
commit()
sync_schema_props_perms('seda_keyword_scheme')



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for OAI-PMH protocol"""
# TODO:  build the rql using syntax tree rather than string concatenation
# (using rqlquery?)


from abc import ABCMeta, abstractmethod

from cubicweb.predicates import is_in_state, relation_possible, score_entity
from cubicweb.view import EntityAdapter
from cubicweb.web.component import Component

from cubes.saem_ref import isodate


def parse_setspec(setspec):
    """Split a setpec string and return <etype>, <relation>, <value> parts."""
    if not setspec:
        raise ValueError('empty setspec')
    parts = setspec.split(':', 2)
    if len(parts) == 3:
        # setspec = etype:relation:value
        return parts
    elif len(parts) == 1:
        # setspec = etype
        return parts[0], None, None
    else:
        raise ValueError('only handling one or three levels set specifiers')


def build_setspec(*specitems):
    """Build a setspec from individual specifier items."""
    return ':'.join(specitems)


class OAIComponent(Component):
    """OAI-PMH protocol component handling set specifiers listing and querying.
    """
    __regid__ = 'oai'
    # maximum number of items for request returning a list of discrete items
    max_result_size = 20
    __setspecs__ = {}

    @classmethod
    def register(cls, specifier):
        """Register a top-level set specifier."""
        cls.__setspecs__[specifier.setkey] = specifier

    def setspecs(self):
        """Yield known set specifier string values"""
        for specifier in self.__setspecs__.itervalues():
            for setkey, description in specifier.setspecs(self._cw):
                yield setkey, description
                for subkey, child in specifier.iteritems():
                    for value, description in child.setspecs(self._cw):
                        yield build_setspec(setkey, subkey, value), description

    def match_setspec(self, setspec, from_date=None, until_date=None,
                      token=None):
        """Return a ResultSet or None and a new token value.

        A result set is returned if `setspec` set specifier matches any of the
        adapter `setspecs`; else it is None. This result set has a size lower
        that class `max_result_size` attribute.

        A new token value is returned (in case a result set is returned, None
        otherwise) if there are more results to fetch from the same query.
        """
        etype, relation, value = parse_setspec(setspec)
        try:
            root = setspec = self.__setspecs__[etype]
        except KeyError:
            raise ValueError('no setspec matching {0}'.format(etype))
        if relation:
            try:
                setspec = root[relation]
            except KeyError:
                raise ValueError('unregistered setspec {0}'.format(setspec))
        setspec_where, args = setspec.setspec_restrictions(value)
        restrictions = [setspec_where]
        # Add dates restrictions.
        if from_date is not None:
            restrictions.append('X {0} >= %(from_date)s'.format(root.date_attr))
            args['from_date'] = from_date
        if until_date is not None:
            restrictions.append('X {0} <= %(until_date)s'.format(root.date_attr))
            args['until_date'] = until_date
        # Add token restriction.
        if token is not None:
            restrictions.append('X eid >= %(offset)s')
            args['offset'] = int(token)
        # Final query.
        qs = ('Any X ORDERBY X LIMIT {0} WHERE '.format(self.max_result_size + 1) +
              ', '.join(restrictions))
        rset = self._cw.execute(qs, args)
        return self.limit_rset(rset)

    def limit_rset(self, rset):
        """Return a result set limited to `max_result_size` along with a token.

        New token is the eid of the last entity in the result set (which is
        not included in the returned result set).
        """
        if len(rset) > self.max_result_size:
            token = str(rset[-1][0])
            rset = rset.limit(self.max_result_size, inplace=True)
        else:
            token = None
        return rset, token

    def match_identifier(self, identifier):
        """Return a result set matching identifier."""
        results = []
        for spec in self.__setspecs__.itervalues():
            qs, args = spec.query_for_identifier(unicode(identifier))
            rset = self._cw.execute(qs, args)
            if rset:
                assert not results, \
                    'ambiguous result for identifier {0}'.format(identifier)
                results.append(rset)
        return results[0] if results else self._cw.empty_rset()


class OAIPMHRecordAdapter(EntityAdapter):
    """Entity adapter representing an OAI-PMH record.

    Concret adapter are implemented for one entity type and define a top-level
    specifier to be returned by `set_definition` method.
    """
    __regid__ = 'IOAIPMHRecord'
    __abstract__ = True
    # the name of the attribute to be used as OAI identifier
    identifier_attribute = 'eid'
    # view identifier to generate metadata of OAI-PMH record
    metadata_view = None
    # date attribute to filter results by from/until request parameters
    date_attr = 'modification_date'
    # indicate that the record got deleted
    deleted = False

    @classmethod
    def __registered__(cls, *args):
        specifier = cls.set_definition()
        if specifier is not None:
            OAIComponent.register(specifier)
        return super(OAIPMHRecordAdapter, cls).__registered__(*args)

    @classmethod
    def set_definition(cls):
        """Return the top-level set specifier for adapter entity type.

        May return None if the adapter should not be registered within
        `OAIComponent`, e.g., if it is only used to build a record from an
        entity queried by its identifier not by a setspec.
        """
        return None

    @property
    def identifier(self):
        """OAI-PMH identifier for adapted entity."""
        return unicode(getattr(self.entity, self.identifier_attribute))

    @property
    def date(self):
        """OAI-PMH date information for adapted entity."""
        return getattr(self.entity, self.date_attr)

    @property
    def datestamp(self):
        """String representation of OAI-PMH date information for adapted entity.
        """
        return isodate(self.date)

    def metadata(self):
        """Return an XML-formatted representation of adapted entity."""
        if self.metadata_view is None:
            raise NotImplementedError(
                'undefined metadata_view attribute for {0}'.format(self.__class__))
        data = self._cw.view(self.metadata_view, w=None,
                             rset=self.entity.as_rset())
        if isinstance(data, unicode):
            # Underlying view may be 'binary' or not.
            data = data.encode('utf-8')
        return data


class OAISetSpec(object):
    """Base class for an OAI-PMH request set specifier."""
    __metaclass__ = ABCMeta

    @abstractmethod
    def setspecs(self, cnx):
        """Yield setspec strings associated with this entry in the the setspec
        hierarchy.
        """

    @abstractmethod
    def setspec_restrictions(self, value=None):
        """Return the WHERE part of the RQL query string and a dict of query
        parameters for this set specifier.
        """


class ETypeOAISetSpec(OAISetSpec, dict):
    """OAI-PMH set specifier matching an entity type."""

    def __init__(self, etype, identifier_attribute, setkey=None,
                 date_attr=OAIPMHRecordAdapter.date_attr):
        """Build a top-level set specifier."""
        super(ETypeOAISetSpec, self).__init__()
        if setkey is None:
            setkey = etype.lower()
        self.etype = etype
        self.identifier_attribute = identifier_attribute
        self.setkey = setkey
        self.date_attr = date_attr

    def __setitem__(self, subset, specifier):
        """Register a new `OAISetSpec` as child of this object."""
        super(ETypeOAISetSpec, self).__setitem__(subset, specifier)
        specifier.__parent__ = self

    def setspecs(self, cnx):
        yield self.setkey, cnx._(self.etype)

    def setspec_restrictions(self, value=None):
        if value is not None:
            raise ValueError('unexpected setspec')
        return 'X is ' + self.etype, {}

    def query_for_identifier(self, identifier):
        """Return an RQL query string and a dict of query parameter for given
        `identifier` within the context of this set specifier.
        """
        qs = 'Any X WHERE X is {etype}, X {attr} %(identifier)s'.format(
            etype=self.etype, attr=self.identifier_attribute)
        return qs, {'identifier': identifier}


class RelatedEntityOAISetSpec(OAISetSpec):
    """OAI-PMH second-level set specifier to match on a relation on the
    top-level entry.
    """

    def __init__(self, rtype, targettype, targetattr,
                 description=None, exclude=()):
        self.rtype = rtype
        self.targettype = targettype
        self.targetattr = targetattr
        self.description = description
        self.exclude = exclude
        self.__parent__ = None

    def setspecs(self, cnx):
        qs = 'DISTINCT String N WHERE X is {etype}, X {attr} N'
        if self.exclude:
            if len(self.exclude) > 1:
                qs += ', NOT X {{attr}} IN ({0})'.format(
                    ','.join(map(repr, self.exclude)))  # pylint: disable=bad-builtin
            else:
                qs += ', NOT X {{attr}} "{0}"'.format(
                    self.exclude[0])
        rset = cnx.execute(qs.format(
            etype=self.targettype, attr=self.targetattr))
        for value, in rset.rows:
            desc = self.description.format(value) if self.description else u''
            yield value, desc

    def setspec_restrictions(self, value):
        baseqs, args = self.__parent__.setspec_restrictions()
        subqs = 'X {rtype} Y, Y {attrname} %(value)s'.format(
            rtype=self.rtype, attrname=self.targetattr)
        qs = ', '.join([baseqs, subqs])
        args['value'] = unicode(value)
        return qs, args


# SAEM-specific implentation ###################################################

class ArkOAIPMHRecordAdapter(OAIPMHRecordAdapter):
    """OAI-PMH Record using the "ark" attribute of entity as identifier.
    """

    __abstract__ = True
    __select__ = (OAIPMHRecordAdapter.__select__ &
                  score_entity(lambda x: hasattr(x, 'ark')))
    identifier_attribute = 'ark'


def in_state(state):
    """Predicate checking for workflowability and matching on `state`.
    """
    return relation_possible('in_state') & is_in_state(state)


def not_in_state(state):
    """Predicate checking for workflowability and not matching on `state`.
    """
    return relation_possible('in_state') & ~is_in_state(state)


class OAIPMHActiveRecordAdapter(ArkOAIPMHRecordAdapter):
    """OAI-PMH Record for "active" entities (not deprecated in our application).

    See http://www.openarchives.org/OAI/2.0/openarchivesprotocol.htm#DeletedRecords
    """

    __abstract__ = True
    __select__ = ArkOAIPMHRecordAdapter.__select__ & not_in_state('deprecated')


class OAIPMHDeletedRecordAdapter(ArkOAIPMHRecordAdapter):
    """OAI-PMH Record for "deleted" entities (deprecated in our application).

    See http://www.openarchives.org/OAI/2.0/openarchivesprotocol.htm#DeletedRecords

    Handle the representation of entity with an ARK which is deleted as an
    OAI-PMH record.
    """

    __select__ = ArkOAIPMHRecordAdapter.__select__ & in_state('deprecated')

    @property
    def datestamp(self):
        """For delete a record, the date and time that it was deleted on.
        """
        trinfo = self.entity.cw_adapt_to('IWorkflowable').latest_trinfo()
        assert trinfo.transition.name == 'deprecate', \
            'unexpected latest transition {0}'.format(trinfo.transition)
        return isodate(trinfo.creation_date)

    deleted = True

    def metadata(self):
        return ''


class PublicETypeOAISetSpec(ETypeOAISetSpec):
    """OAI-PMH set specifier matching an entity type and entities in
    "published" or "deprecated" state.
    """

    def setspec_restrictions(self, value=None):
        """Add workflow state restrictions."""
        base, args = super(PublicETypeOAISetSpec, self).setspec_restrictions(value)
        restrictions = ', '.join([base, 'X in_state ST, NOT ST name "draft"'])
        return restrictions, args

    def query_for_identifier(self, identifier):
        qs, args = super(PublicETypeOAISetSpec, self).query_for_identifier(
            identifier)
        qs += ', X in_state ST, NOT ST name "draft"'
        return qs, args



from cubicweb.schema import ETYPE_NAME_MAP
ETYPE_NAME_MAP['ProfileArchive'] = 'ProfileArchiveObject'
sql('ALTER TABLE cw_ProfileArchive add column cw_seda_parent integer')
sql('UPDATE cw_ProfileArchive SET cw_seda_parent=cw_seda_archive_profile')
drop_relation_type('seda_archive_profile')
rename_entity_type('ProfileArchive', 'ProfileArchiveObject')
sql("UPDATE entities SET type='ProfileArchiveObject' WHERE type='ProfileArchive';")

for etype in ('FileTypeCode', 'CharacterSetCode', 'DocumentTypeCode'):
    rename_entity_type(etype, 'SEDA' + etype)



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for exporting agents to eac"""

from lxml import etree

from logilab.common.date import ustrftime
from cubicweb.predicates import is_instance

from cubes.saem_ref import (cwuri_url, eactype_mapping, eacaddress_mapping,
                            eacmaintenancetype_mapping)
from cubes.saem_ref.entities import AbstractXmlAdapter


class AgentEACAdapter(AbstractXmlAdapter):
    __regid__ = 'EAC-CPF'
    __select__ = is_instance('Agent')

    namespaces = {
        None: u'urn:isbn:1-931666-33-4',
        u'xsi': u'http://www.w3.org/2001/XMLSchema-instance',
        u'xlink': u'http://www.w3.org/1999/xlink',
    }
    datetime_fmt = '%Y-%m-%dT%H:%M:%S'

    @property
    def file_name(self):
        """Return a file name for the dump."""
        return u'{0}.xml'.format(self.entity.dc_title())

    def dump(self):
        """Return an XML string representing the given agent using the EAC-CPF schema."""
        # Keep related activities since they are used multiple times
        self.activities = sorted(self.entity.reverse_generated, key=lambda x: x.start, reverse=True)
        # Root element
        eac_cpf_elt = self.element('eac-cpf', attributes={
            'xsi:schemaLocation': ('urn:isbn:1-931666-33-4 '
                                   'http://eac.staatsbibliothek-berlin.de/schema/cpf.xsd')
        })
        # Top elements: control & cpfDescription
        self.control_element(eac_cpf_elt)
        self.cpfdescription_element(eac_cpf_elt)
        tree = etree.ElementTree(eac_cpf_elt)
        return etree.tostring(tree, xml_declaration=True, encoding='utf-8', pretty_print=True)

    def control_element(self, eac_cpf_elt):
        control_elt = self.element('control', parent=eac_cpf_elt)
        self.element('recordId', parent=control_elt, text=self.entity.ark)
        self.maintenance_status_element(control_elt)
        self.publication_status_element(control_elt)
        self.maintenance_agency_element(control_elt)
        self.language_declaration_element(control_elt)
        self.maintenance_history_element(control_elt)
        self.sources_element(control_elt)

    def cpfdescription_element(self, eac_cpf_elt):
        cpfdescription_elt = self.element('cpfDescription', parent=eac_cpf_elt)
        self.identity_element(cpfdescription_elt)
        self.description_element(cpfdescription_elt)
        self.relations_element(cpfdescription_elt)

    def maintenance_status_element(self, control_elt):
        if any(activity.type == 'modify' for activity in self.activities):
            status = 'revised'
        else:
            status = 'new'
        self.element('maintenanceStatus', parent=control_elt, text=status)

    def publication_status_element(self, control_elt):
        status = {'draft': 'inProcess', 'published': 'approved'}.get(
            self.entity.cw_adapt_to('IWorkflowable').state, 'inProcess')
        self.element('publicationStatus', parent=control_elt, text=status)

    def maintenance_agency_element(self, control_elt):
        agency_elt = self.element('maintenanceAgency', parent=control_elt)
        self.element('agencyName', parent=agency_elt)

    def language_declaration_element(self, control_elt):
        lang_decl_elt = self.element('languageDeclaration', parent=control_elt)
        self.element('language', parent=lang_decl_elt,
                     attributes={'languageCode': 'fre'}, text=u'français')
        self.element('script', parent=lang_decl_elt, attributes={'scriptCode': 'Latn'},
                     text='latin')

    def maintenance_history_element(self, control_elt):
        history_elt = self.element('maintenanceHistory', parent=control_elt)
        for activity in self.activities:
            self.maintenance_event_element(activity, history_elt)

    def sources_element(self, control_elt):
        sources_elt = self.element('sources')
        for eac_source in self.entity.reverse_source_agent:
            self.source_element(eac_source, sources_elt)
        if len(sources_elt):
            control_elt.append(sources_elt)

    def identity_element(self, cpfdescription_elt):
        identity_elt = self.element('identity', parent=cpfdescription_elt)
        self._elt_text_from_attr('entityId', self.entity, 'isni', parent=identity_elt)
        type_mapping = {v: k for k, v in eactype_mapping.items()}
        eac_type = type_mapping.get(self.entity.kind)
        self.element('entityType', parent=identity_elt, text=eac_type)
        name_entry = self.element('nameEntry', parent=identity_elt)  # XXX: authorizedForm
        self.element('part', parent=name_entry, text=self.entity.name)

    def description_element(self, cpfdescription_elt):
        agent = self.entity
        description_elt = self.element('description', parent=cpfdescription_elt)
        self.exist_dates_element(description_elt)
        for place in agent.reverse_place_agent:
            self.place_element(place, description_elt)
        for function in agent.reverse_function_agent:
            self.function_element(function, description_elt)
        for legal_status in agent.reverse_legal_status_agent:
            self.legal_status_element(legal_status, description_elt)
        for mandate in agent.reverse_mandate_agent:
            self.mandate_element(mandate, description_elt)
        for structure in agent.reverse_structure_agent:
            self.structure_element(structure, description_elt)
        for history in agent.reverse_history_agent:
            self.bioghist_element(history, description_elt)

    def relations_element(self, cpfdescription_elt):
        relations_elt = self.element('relations')
        for rtype, rel_rtype, eac_rtype in [
            ('reverse_hierarchical_child', 'hierarchical_parent', 'hierarchical-parent'),
            ('reverse_hierarchical_parent', 'hierarchical_child', 'hierarchical-child'),
            ('reverse_chronological_successor', 'chronological_predecessor', 'temporal-earlier'),
            ('reverse_chronological_predecessor', 'chronological_successor', 'temporal-after'),
            ('reverse_association_to', 'association_from', 'associative'),
            ('reverse_association_from', 'association_to', 'associative')
        ]:
            for relation in getattr(self.entity, rtype):
                self.cpfrelation_element(relation, relations_elt, cw_rtype=rel_rtype,
                                         eac_rtype=eac_rtype)
        for resource_relation in self.entity.reverse_resource_relation_agent:
            self.resource_relation_element(resource_relation, relations_elt)
        if len(relations_elt):
            cpfdescription_elt.append(relations_elt)

    def maintenance_event_element(self, activity, history_elt):
        event_elt = self.element('maintenanceEvent', parent=history_elt)
        type_mapping = {v: k for k, v in eacmaintenancetype_mapping.items()}
        activity_type = type_mapping.get(activity.type, 'created')
        self.element('eventType', parent=event_elt, text=activity_type)
        self.element('eventDateTime', parent=event_elt,
                     attributes={'standardDateTime': ustrftime(activity.start,
                                                               fmt=self.datetime_fmt)},
                     text=ustrftime(activity.start, fmt=self.datetime_fmt))
        ass_agent = activity.associated_with
        if ass_agent:
            self.agent_element(ass_agent[0], event_elt)
        else:  # These tags must be present, even if name is empty
            agent_type_elt = self.element('agentType', text='machine')
            event_elt.append(agent_type_elt)
            event_elt.append(self.element('agent'))
        self._elt_text_from_attr('eventDescription', activity, 'description', parent=event_elt)

    def source_element(self, eac_source, sources_elt):
        url = eac_source.url
        attributes = {'xlink:href': url, 'xlink:type': 'simple'} if url else None
        source_elt = self.element('source', attributes=attributes, parent=sources_elt)
        self.element('sourceEntry', parent=source_elt, text=eac_source.title)
        self._eac_inner_paragraph_element('descriptiveNote', eac_source.description,
                                          parent=source_elt)

    def exist_dates_element(self, description_elt):
        date_range = self._eac_date_range_xml_elt(self.entity.start_date, self.entity.end_date)
        if date_range is not None:
            exist_dates = self.element('existDates', parent=description_elt)
            exist_dates.append(date_range)

    def place_element(self, place, description_elt):
        place_elt = self.element('place', parent=description_elt)
        for attr, eac_name in [('role', 'placeRole'), ('name', 'placeEntry')]:
            eac_elt = self._elt_text_from_attr(eac_name, place, attr, parent=place_elt)
            if eac_elt is not None and eac_name == 'placeEntry' and place.equivalent_concept:
                eac_elt.attrib['vocabularySource'] = cwuri_url(place.equivalent_concept[0])
        for address in place.place_address:
            self.address_element(address, place_elt)

    def function_element(self, function, description_elt):
        function_elt = self.element('function', parent=description_elt)
        term_elt = self._elt_text_from_attr('term', function, 'name', parent=function_elt)
        if term_elt is not None and function.equivalent_concept:
            term_elt.attrib['vocabularySource'] = cwuri_url(function.equivalent_concept[0])
        self._eac_inner_paragraph_element('descriptiveNote', function.description,
                                          parent=function_elt)

    def legal_status_element(self, legal_status, description_elt):
        legal_status_elt = self.element('legalStatus', parent=description_elt)
        self._elt_text_from_attr('term', legal_status, 'term', parent=legal_status_elt)
        self._eac_inner_paragraph_element('descriptiveNote', legal_status.description,
                                          parent=legal_status_elt)

    def mandate_element(self, mandate, description_elt):
        mandate_elt = self.element('mandate', parent=description_elt)
        term_elt = self._elt_text_from_attr('term', mandate, 'term', parent=mandate_elt)
        if term_elt is not None and mandate.equivalent_concept:
            term_elt.attrib['vocabularySource'] = cwuri_url(mandate.equivalent_concept[0])
        self._eac_inner_paragraph_element('descriptiveNote', mandate.description,
                                          parent=mandate_elt)

    def structure_element(self, structure, description_elt):
        self._eac_inner_paragraph_element('structureOrGenealogy', structure.description,
                                          parent=description_elt)

    def bioghist_element(self, history, description_elt):
        self._eac_inner_paragraph_element('biogHist', history.text, parent=description_elt)

    def cpfrelation_element(self, relation, relations_elt, cw_rtype=None,
                            eac_rtype='associative'):
        related = getattr(relation, cw_rtype, [None])[0]
        assert related is not None
        relation_elt = self.element('cpfRelation',
                                    attributes={'cpfRelationType': eac_rtype,
                                                'xlink:href': related.absolute_url(),
                                                'xlink:type': 'simple'},
                                    parent=relations_elt)
        if related.cw_etype != 'ExternalUri':
            self.element('relationEntry', parent=relation_elt, text=related.name)
        self._eac_date_range_xml_elt(getattr(relation, 'start_date', None),
                                     getattr(relation, 'end_date', None),
                                     parent=relation_elt)
        self._eac_inner_paragraph_element('descriptiveNote', relation.description,
                                          parent=relation_elt)

    def resource_relation_element(self, resource_relation, relations_elt):
        resource = resource_relation.resource_relation_resource[0]
        res_rel_elt = self.element('resourceRelation',
                                   attributes={'resourceRelationType': resource_relation.agent_role,
                                               'xlink:href': resource.uri,
                                               'xlink:role': resource_relation.resource_role,
                                               'xlink:type': 'simple'},
                                   parent=relations_elt)
        self._eac_date_range_xml_elt(resource_relation.start_date, resource_relation.end_date,
                                     parent=res_rel_elt)

    def agent_element(self, agent, maintenance_event_elt):
        self.element('agentType', maintenance_event_elt, text='human')
        self.element('agent', maintenance_event_elt, text=agent.name)

    def address_element(self, address, place_elt):
        address_elt = self.element('address', parent=place_elt)
        for eac_name, attr in eacaddress_mapping:
            self._elt_text_from_attr('addressLine', address, attr,
                                     attributes={'localType': eac_name}, parent=address_elt)

    #
    # helper methods for lxml
    #

    def _elt_text_from_attr(self, tag_name, entity, attr_name, parent=None, attributes=None):
        """Return an lxml `Element` whose text is the value of the given attribute on the given
        entity.

        If this element is not empty and if ``parent`` is not ``None``, the element will also be
        inserted in the parent XML element.

        If ``attributes`` is not ``None``, these attributes will be added to the returned element.

        Return ``None`` if element is empty.
        """
        value = getattr(entity, attr_name)
        if value is not None:
            elt = self.element(tag_name, parent=parent, attributes=attributes, text=value)
            return elt

    def _eac_date_range_xml_elt(self, start_date, end_date, parent=None):
        """Return an EAC lxml ``'dateRange'`` ``Element`` with the given boundaries."""
        if not start_date and not end_date:
            return
        date_range = self.element('dateRange', parent=parent)
        for dt, eac_name in [(start_date, 'fromDate'), (end_date, 'toDate')]:
            if not dt:
                continue
            self.element(eac_name, parent=date_range,
                         attributes={'standardDate': dt.isoformat()}, text=dt.isoformat())
        return date_range

    def _eac_inner_paragraph_element(self, tag, text, parent):
        """Return an EAC lxml ``Element`` having the given tag and enclosing a ``<p>`` element with
        the given text."""
        if text:
            descr_elt = self.element(tag, parent=parent)
            self.element('p', parent=descr_elt, text=text)



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tests for agent entities"""

import datetime

from mock import MagicMock, patch

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.rdfio import RDFLibRDFGraph

import testutils


class AgentExportTC(CubicWebTC, testutils.XmlTestMixin):
    """Test case for agent exports"""

    @patch('cubes.saem_ref.hooks.datetime')
    def setup_database(self, mock_dt):
        mock_dt.utcnow.return_value = datetime.datetime(2016, 2, 1)
        with self.admin_access.client_cnx() as cnx:
            org = cnx.find('AgentKind', name=u'authority').one()
            person = cnx.find('AgentKind', name=u'person').one()
            unknown = cnx.find('AgentKind', name=u'unknown-agent-kind').one()
            archival_role = cnx.find('ArchivalRole', name=u'archival').one()
            control_role = cnx.find('ArchivalRole', name=u'control').one()
            deposit_role = cnx.find('ArchivalRole', name=u'deposit').one()
            work_address2 = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                              postalcode=u'31400', city=u'Toulouse')
            work_address3 = cnx.create_entity('PostalAddress', street=u'104 bd L.-A. Blanqui',
                                              postalcode=u'75013', city=u'Paris')
            home_address = cnx.create_entity('PostalAddress', street=u'Place du Capitole',
                                             postalcode=u'31000', city=u'Toulouse')
            agent1 = cnx.create_entity('Agent', name=u'Alice', agent_kind=person,
                                       archival_role=[deposit_role])
            agent2 = cnx.create_entity('Agent', name=u'Super Service', agent_kind=org,
                                       contact_point=agent1,
                                       archival_role=[archival_role],
                                       reverse_archival_agent=agent1)
            cnx.create_entity('AgentPlace', role=u'work', place_agent=agent2,
                              place_address=work_address2)
            agent3 = cnx.create_entity('Agent', name=u'Charlie', agent_kind=person, ark=u'1234',
                                       archival_role=[archival_role, control_role])
            cnx.create_entity('AgentPlace', role=u'work', place_agent=agent3,
                              place_address=work_address3)
            cnx.create_entity('AgentPlace', role=u'home', place_agent=agent3,
                              place_address=home_address)
            cnx.create_entity('ChronologicalRelation', chronological_predecessor=agent2,
                              chronological_successor=agent3)
            gironde = cnx.create_entity('Agent', name=u'Gironde. Conseil général', agent_kind=org,
                                        ark=u'FRAD033_EAC_00001',
                                        start_date=datetime.date(1800, 1, 1),
                                        end_date=datetime.date(2099, 1, 1),
                                        isni=u'22330001300016',
                                        archival_role=[archival_role, control_role])
            delphine = cnx.create_entity('Agent', name=u'Delphine Jamet', agent_kind=person)
            adm_dir = cnx.create_entity('Agent', name=u"Gironde. Conseil général. Direction de "
                                        u"l'administration et de la sécurité juridique",
                                        agent_kind=unknown)
            cg32 = cnx.create_entity('Agent', name=u'CG32', agent_kind=unknown)
            place_uri = cnx.create_entity('ExternalUri',
                                          uri=u'http://catalogue.bnf.fr/ark:/12148/cb152418385',
                                          cwuri=u'http://catalogue.bnf.fr/ark:/12148/cb152418385')
            agent_x = cnx.create_entity('ExternalUri', uri=u'agent-x', cwuri=u'agent-x')
            social_action_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200',
                cwuri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200')
            environment_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074',
                cwuri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074')
            resource_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N',
                cwuri=u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N')
            cnx.create_entity('EACSource', title=u'1. Ouvrages imprimés...',
                              description=u'des bouquins', source_agent=gironde)
            cnx.create_entity('EACSource', title=u'Site des Archives départementales de la Gironde',
                              url=u'http://archives.gironde.fr', source_agent=gironde)
            cnx.create_entity('Activity', type=u'create',
                              start=datetime.datetime(2013, 4, 24, 5, 34, 41),
                              end=datetime.datetime(2013, 4, 24, 5, 34, 41),
                              description=u'bla bla bla', generated=gironde)
            cnx.create_entity('Activity', type=u'modify', associated_with=delphine,
                              start=datetime.datetime(2015, 1, 15, 7, 16, 33),
                              end=datetime.datetime(2015, 1, 15, 7, 16, 33),
                              generated=gironde)
            gironde_address = cnx.create_entity('PostalAddress',
                                                street=u'1 Esplanade Charles de Gaulle',
                                                postalcode=u'33074', city=u'Bordeaux Cedex')
            cnx.create_entity('AgentPlace', role=u'siege', name=u'Bordeaux (Gironde, France)',
                              place_agent=gironde, place_address=gironde_address,
                              equivalent_concept=place_uri)
            cnx.create_entity('AgentPlace', role=u'domicile', name=u'Toulouse (France)',
                              place_agent=gironde)
            cnx.create_entity('AgentPlace', role=u'dodo', name=u'Lit',
                              place_agent=gironde)
            cnx.create_entity('LegalStatus', term=u'Collectivité territoriale',
                              start_date=datetime.date(1234, 1, 1),
                              end_date=datetime.date(3000, 1, 1),
                              description=u'Description du statut',
                              legal_status_agent=gironde)
            cnx.create_entity('Mandate', term=u'1. Constitutions françaises',
                              description=u'Description du mandat',
                              mandate_agent=gironde)
            cnx.create_entity('History',
                              text=u'La loi du 22 décembre 1789, en divisant ...',
                              history_agent=gironde)
            cnx.create_entity('Structure',
                              description=u'Pour accomplir ses missions ...',
                              structure_agent=gironde)
            cnx.create_entity('AgentFunction',
                              description=u'Quatre grands domaines de compétence...',
                              function_agent=gironde)
            cnx.create_entity('AgentFunction', name=u'action sociale',
                              description=u'1. Solidarité\nblablabla.',
                              function_agent=gironde, equivalent_concept=social_action_uri)
            cnx.create_entity('AgentFunction', name=u'environnement',
                              function_agent=gironde, equivalent_concept=environment_uri)
            cnx.create_entity('HierarchicalRelation', hierarchical_parent=adm_dir,
                              hierarchical_child=gironde, description=u'Coucou',
                              start_date=datetime.date(2008, 1, 1),
                              end_date=datetime.date(2099, 1, 1))
            cnx.create_entity('ChronologicalRelation', chronological_predecessor=cg32,
                              chronological_successor=gironde)
            cnx.create_entity('AssociationRelation', association_from=gironde,
                              association_to=agent_x)
            cnx.create_entity('EACResourceRelation', agent_role=u'creatorOf',
                              resource_role=u"Fonds d'archives",
                              start_date=datetime.date(1673, 1, 1),
                              end_date=datetime.date(1963, 1, 1),
                              resource_relation_resource=resource_uri,
                              resource_relation_agent=gironde)
            cnx.commit()
            gironde.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            self.agent1_eid = agent1.eid
            self.agent2_eid = agent2.eid
            self.agent3_eid = agent3.eid
            self.gironde_eid = gironde.eid
            self.adm_dir_eid = adm_dir.eid
            self.cg32_eid = cg32.eid
            self.agent_x_eid = agent_x.eid
            self.work_address2_eid = work_address2.eid
            self.work_address3_eid = work_address3.eid
            self.home_address_eid = home_address.eid

    def assertItemsIn(self, items, container):
        """Check that elements of `items` are in `container`."""
        for item in items:
            self.assertIn(item, container)

    def assertItemsNotIn(self, items, container):
        """Check that elements of `items` are not in `container`."""
        for item in items:
            self.assertNotIn(item, container)

    def test_agent_rdf_primary_adapter(self):
        role_uri = u'http://www.w3.org/2006/vcard/ns#role'
        type_uri = u'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'
        person_uri = u'http://xmlns.com/foaf/0.1/Person'
        org_uri = u'http://xmlns.com/foaf/0.1/Organization'
        contact_uri = u'http://schema.org/contactPoint'
        name_uri = u'http://xmlns.com/foaf/0.1/name'
        address_uri = u'http://www.w3.org/2006/vcard/ns#hasAddress'
        archival_agent_uri = u'http://www.logilab.org/saem/0#archivalAgent'
        replaced_uri = u'http://purl.org/dc/terms/isReplacedBy'
        replaces_uri = u'http://purl.org/dc/terms/replaces'
        with self.admin_access.client_cnx() as cnx:
            agent1 = cnx.entity_from_eid(self.agent1_eid)
            agent2 = cnx.entity_from_eid(self.agent2_eid)
            agent3 = cnx.entity_from_eid(self.agent3_eid)
            work_address2 = cnx.entity_from_eid(self.work_address2_eid)
            work_address3 = cnx.entity_from_eid(self.work_address3_eid)
            home_address = cnx.entity_from_eid(self.home_address_eid)
            # No archival role / no address
            graph = RDFLibRDFGraph()
            agent1.cw_adapt_to('RDFPrimary').fill(graph)
            triples = list(graph.triples())
            self.assertItemsIn([(agent1.absolute_url(), type_uri, person_uri),
                                (agent1.absolute_url(), name_uri, u'Alice')],
                               triples)
            self.assertItemsIn([(agent1.absolute_url(), archival_agent_uri, agent2.absolute_url())],
                               triples)
            self.assertItemsNotIn(
                [(agent1.absolute_url(), role_uri, u'archival'),
                 (agent1.absolute_url(), role_uri, u'control'),
                 (agent1.absolute_url(), address_uri, work_address2.absolute_url()),
                 (agent1.absolute_url(), address_uri, work_address3.absolute_url()),
                 (agent1.absolute_url(), address_uri, home_address.absolute_url()),
                 (agent2.absolute_url(), name_uri, u'Super Service'),
                 ], triples)
            # One archival role / one address
            graph = RDFLibRDFGraph()
            agent2.cw_adapt_to('RDFPrimary').fill(graph)
            triples = list(graph.triples())
            self.assertItemsIn([(agent2.absolute_url(), type_uri, org_uri),
                                (agent2.absolute_url(), name_uri, u'Super Service'),
                                (agent2.absolute_url(), role_uri, u'archival'),
                                (agent2.absolute_url(), address_uri,
                                 work_address2.absolute_url()),
                                (work_address2.absolute_url(), role_uri, 'work')],
                               triples)
            # agent1 in the graph as contact point of agent2
            self.assertItemsIn([(agent2.absolute_url(), contact_uri, agent1.absolute_url()),
                                (agent1.absolute_url(), type_uri, person_uri),
                                (agent1.absolute_url(), name_uri, u'Alice')],
                               triples)
            self.assertItemsNotIn(
                [(agent2.absolute_url(), role_uri, u'control'),
                 (agent2.absolute_url(), address_uri, work_address3.absolute_url()),
                 (work_address3.absolute_url(), role_uri, 'work'),
                 (agent2.absolute_url(), address_uri, home_address.absolute_url())],
                triples)
            # Two archival roles
            graph = RDFLibRDFGraph()
            agent3.cw_adapt_to('RDFPrimary').fill(graph)
            triples = list(graph.triples())
            self.assertItemsIn(
                [(agent3.absolute_url(), type_uri, person_uri),
                 (agent3.absolute_url(), name_uri, u'Charlie'),
                 (agent3.absolute_url(), role_uri, u'archival'),
                 (agent3.absolute_url(), role_uri, u'control'),
                 (agent3.absolute_url(), address_uri, work_address3.absolute_url()),
                 (work_address3.absolute_url(), role_uri, 'work'),
                 (agent3.absolute_url(), address_uri, home_address.absolute_url()),
                 (home_address.absolute_url(), role_uri, 'home')],
                triples)
            self.assertNotIn(
                (agent3.absolute_url(), address_uri, work_address2.absolute_url()),
                triples)
            # agent1 replaced by agent2 and agent2 replaced by agent3
            self.assertItemsIn(
                [(agent2.absolute_url(), replaced_uri, agent3.absolute_url()),
                 (agent3.absolute_url(), replaces_uri, agent2.absolute_url())],
                triples)

    def test_agent_eac_export(self):
        """Check that an agent is correctly exported into a valid EAC-CPF XML file."""
        # Given an agent (created in setup_database), export it to EAC-CPF
        with self.admin_access.client_cnx() as cnx:
            agent3 = cnx.entity_from_eid(self.agent3_eid)
            agent2 = cnx.entity_from_eid(self.agent2_eid)
            agent2.absolute_url = MagicMock(return_value=u'http://www.example.org/agent2')
            agent3_eac = agent3.cw_adapt_to('EAC-CPF').dump()
        # Then check that output XML is as expected
        with open(self.datapath('EAC/agent3_export.xml')) as f:
            expected_eac = f.read()
        self.assertXmlEqual(agent3_eac, expected_eac)
        self.assertXmlValid(agent3_eac, self.datapath('cpf.xsd'))
        # Same for a more complete example
        with self.admin_access.client_cnx() as cnx:
            adm_dir = cnx.entity_from_eid(self.adm_dir_eid)
            adm_dir.absolute_url = MagicMock(return_value=u'http://www.example.org/adm_dir')
            cg32 = cnx.entity_from_eid(self.cg32_eid)
            cg32.absolute_url = MagicMock(return_value=u'http://www.example.org/cg32')
            agent_x = cnx.entity_from_eid(self.agent_x_eid)
            agent_x.absolute_url = MagicMock(return_value=u'http://www.example.org/agent_x')
            gironde = cnx.entity_from_eid(self.gironde_eid)
            gironde_eac = gironde.cw_adapt_to('EAC-CPF').dump()
        with open(self.datapath('EAC/FRAD033_EAC_00001_simplified_export.xml')) as f:
            expected_eac = f.read()
        self.assertXmlEqual(gironde_eac, expected_eac)
        self.assertXmlValid(gironde_eac, self.datapath('cpf.xsd'))


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()
